import logging, os
from pathlib import Path
import re
from typing import Any, List, Dict, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr

import astropy.units
from casacore import tables

from .table_query import open_query, open_table_ro


CASACORE_TO_PD_TIME_CORRECTION = 3506716800.0
SECS_IN_DAY = 86400


def table_exists(path: str) -> bool:
    return tables.tableexists(path)


def convert_casacore_time(rawtimes: np.ndarray, convert_to_datetime=True) -> np.ndarray:
    """
    Read time columns to datetime format
    pandas datetimes are referenced against a 0 of 1970-01-01
    CASA's modified julian day reference time is (of course) 1858-11-17

    This requires a correction of 3506716800 seconds which is hardcoded to save time

    :param rawtimes: times in casacore ref
    :return: times converted to pandas reference
    """

    if convert_to_datetime:
        return pd.to_datetime(
            np.array(rawtimes) - CASACORE_TO_PD_TIME_CORRECTION, unit="s"
        ).values
    else:
        return np.array(rawtimes) - CASACORE_TO_PD_TIME_CORRECTION
    # dt = pd.to_datetime(np.atleast_1d(rawtimes) - correction, unit='s').values
    # if len(np.array(rawtimes).shape) == 0: dt = dt[0]
    # return dt


def convert_mjd_time(rawtimes: np.ndarray) -> np.ndarray:
    """Different time conversion needed for the MJD col of EPHEM{i}_*.tab
    files (only, as far as I've seen)

    :param rawtimes: MJD times for example from the MJD col of ephemerides tables
    :return: times converted to pandas reference and datetime type
    """
    return pd.to_datetime(
        rawtimes * SECS_IN_DAY - CASACORE_TO_PD_TIME_CORRECTION, unit="s"
    ).values


def extract_table_attributes(infile: str) -> Dict[str, Dict]:
    """
    return a dictionary of table attributes created from MS keywords and column descriptions
    """
    with open_table_ro(infile) as tb_tool:
        kwd = tb_tool.getkeywords()
        attrs = dict([(kk, kwd[kk]) for kk in kwd if kk not in os.listdir(infile)])
        cols = tb_tool.colnames()
        column_descriptions = {}
        for col in cols:
            column_descriptions[col] = tb_tool.getcoldesc(col)
        attrs["column_descriptions"] = column_descriptions
        attrs["info"] = tb_tool.info()

    return attrs


def add_units_measures(
    mvars: Dict[str, xr.DataArray], cc_attrs: Dict[str, Any]
) -> Dict[str, xr.DataArray]:
    """
    Add attributes with units and measure metainfo to the variables passed in the input dictionary

    :param mvars: data variables where to populate units
    :param cc_attrs: dictionary with casacore table attributes (from extract_table_attributes)
    :return: variables with units added in their attributes
    """
    col_descrs = cc_attrs["column_descriptions"]
    # TODO: Should probably loop the other way around, over mvars
    for col in col_descrs:
        var_name = col.lower()
        if var_name in mvars and "keywords" in col_descrs[col]:
            if "QuantumUnits" in col_descrs[col]["keywords"]:
                cc_units = col_descrs[col]["keywords"]["QuantumUnits"]
                if not isinstance(cc_units, list) or not cc_units:
                    logging.warning(
                        f"Invalid units found for column/variable {col}: {cc_units}"
                    )
                mvars[var_name].attrs["units"] = cc_units[0]
                try:
                    astropy.units.Unit(cc_units[0])
                except Exception as exc:
                    logging.warning(
                        f"Unsupported units found for column/variable {col}: "
                        f"{cc_units}. Cannot create an astropy.units.Units object from it: {exc}"
                    )

            if "MEASINFO" in col_descrs[col]["keywords"]:
                cc_meas = col_descrs[col]["keywords"]["MEASINFO"]
                mvars[var_name].attrs["measure"] = {"type": cc_meas["type"]}
                # VarRefCol used for several cols of:
                # - SPECTRAL_WINDOW (MEAS_FREG_REF, in MSv2 std)
                # - FIELD  (PhseDir_Ref, DelayDir_Ref, RefDir_Ref, not in MSv2 std)
                # - POINTING - to be split
                if "VarRefCol" not in cc_meas:
                    mvars[var_name].attrs["measure"]["ref_frame"] = cc_meas["Ref"]
                else:
                    mvars[var_name].attrs["measure"]["ref_frame_data_var"] = cc_meas[
                        "VarRefCol"
                    ]
                    if "TabRefTypes" in cc_meas:
                        mvars[var_name].attrs["measure"].update(
                            {
                                "ref_frame_types": list(cc_meas["TabRefTypes"]),
                                "ref_frame_codes": list(cc_meas["TabRefCodes"]),
                            }
                        )

    return mvars


def make_freq_attrs(spw_xds: xr.Dataset, spw_id: int) -> Dict[str, Any]:
    """Grab the units/measure metainfo for the xds.freq dimension of a
    parttion from the SPECTRAL_WINDOW subtable CTDS attributes.

    Has to read xds_spw.meas_freq_ref and use it as index in the CTDS
    'VarRefCol' attrs of CHAN_FREQ and REF_FREQUENCY to give a
    reference frame to xds_spw.ref_frequency and xds_spw.chan_freq
    (then the ref frame from the second will be pulled to
    xds.freq.attrs)

    :param spw_xds: (metainfo) SPECTRAL_WINDOW xds
    :param spw_id: SPW id of a partition
    :return: attributes (units/measure) for the freq dim of a partition

    """
    fallback_TabRefTypes = [
        "REST",
        "LSRK",
        "LSRD",
        "BARY",
        "GEO",
        "TOPO",
        "GALACTO",
        "LGROUP",
        "CMB",
    ]

    ctds_cols = spw_xds.attrs["other"]["msv2"]["ctds_attrs"]["column_descriptions"]
    cfreq = ctds_cols["CHAN_FREQ"]

    cf_attrs = spw_xds.chan_freq.attrs
    if "MEASINFO" in cfreq["keywords"] and "VarRefCol" in cfreq["keywords"]["MEASINFO"]:
        fattrs = cfreq["keywords"]["MEASINFO"]
        var_ref_col = fattrs["VarRefCol"]
        # This should point to the SPW/MEAS_FREQ_REF col
        meas_freq_ref_idx = spw_xds.data_vars[var_ref_col.lower()].values[spw_id]

        if "TabRefCodes" not in fattrs or "TabRefTypes" not in fattrs:
            # Datasets like vla/ic2233_1.ms say "VarRefCol" but "TabRefTypes" is missing
            ref_frame = fallback_TabRefTypes[meas_freq_ref_idx]
        else:
            ref_type_code = fattrs["TabRefCodes"][meas_freq_ref_idx]
            ref_frame = fattrs["TabRefTypes"][ref_type_code]

        cf_attrs["measure"] = {
            "type": fattrs["type"],
            "ref_frame": ref_frame,
        }

        # Also set the 'VarRefCol' for CHAN_FREQ and REF_FREQUENCEY
        spw_xds.data_vars["chan_freq"].attrs.update(cf_attrs)
        spw_xds.data_vars["ref_frequency"].attrs.update(cf_attrs)

    return cf_attrs


def get_pad_nan(col: np.ndarray) -> np.ndarray:
    """
    Produce a padding/nan value appropriate for a data column
    (for when we need to pad data vars coming from columns with rows of
    variable size array values)

    Parameters
    ----------
    col : np.ndarray
        data being loaded from a table column
    """
    # This is causing frequent warnings for integers. Cast of nan to "int nan"
    # produces -2147483648 but also seems to trigger a
    # "RuntimeWarning: invalid value encountered in cast" (new in numpy>=1.24)
    policy = "warn"
    col_type = np.array(col).dtype
    if np.issubdtype(col_type, np.integer):
        policy = "ignore"
    with np.errstate(invalid=policy):
        pad_nan = np.array([np.nan]).astype(col_type)[0]

    return pad_nan


def redimension_ms_subtable(xds: xr.Dataset, subt_name: str) -> xr.Dataset:
    """Expand a MeasurementSet subtable xds from single dimension (row)
    to multiple dimensions (such as (source_id, time, spectral_window)

    WIP: only works for source, experimenting

    :param xds: dataset to change the dimensions
    :param subt_name: subtable name (SOURCE, etc.)
    :return: transformed xds with data dimensions representing the MS subtable key
    (one dimension for every columns)
    """
    subt_key_cols = {
        "DOPPLER": ["doppler_id", "source_id"],
        "FREQ_OFFSET": [
            "antenna1",
            "antenna2",
            "feed_id",
            "spectral_window_id",
            "time",
        ],
        "SOURCE": ["source_id", "time", "spectral_window_id"],
        "SYSCAL": ["antenna_id", "feed_id", "spectral_window_id", "time"],
        "WEATHER": ["antenna_id", "time"],
        # added tables (MSv3 but not preent in MSv2). Build it from "EPHEMi_... tables
        # Not clear what to do about 'time' var/dim:  , "time"],
        "EPHEMERIDES": ["ephemeris_row_id", "ephemeris_id"],
    }
    key_dims = subt_key_cols[subt_name]

    rxds = xds.copy()
    try:
        with np.errstate(invalid="ignore"):
            rxds = rxds.set_index(row=key_dims).unstack("row").transpose(*key_dims, ...)
        # unstack changes type to float when it needs to introduce NaNs, so
        # we need to reset to the original type.
        for var in rxds.data_vars:
            if rxds[var].dtype != xds[var].dtype:
                rxds[var] = rxds[var].astype(xds[var].dtype)
    except Exception as exc:
        logging.warning(
            f"Cannot expand rows to {key_dims}, possibly duplicate values in those coordinates. Exception: {exc}"
        )
        rxds = xds.copy()

    return rxds


def is_ephem_subtable(tname: str) -> bool:
    return "EPHEM" in tname and Path(tname).name.startswith("EPHEM")


def add_ephemeris_vars(tname: str, xds: xr.Dataset) -> xr.Dataset:
    fname = Path(tname).name
    pattern = r"EPHEM(\d+)_"
    match = re.match(pattern, fname)
    if match:
        ephem_id = match.group(1)
    else:
        ephem_id = 0

    xds["ephemeris_id"] = np.uint32(ephem_id) * xr.ones_like(
        xds["mjd"], dtype=np.uint32
    )
    xds = xds.rename({"mjd": "time"})
    xds["ephemeris_row_id"] = (
        xr.zeros_like(xds["time"], dtype=np.uint32) + xds["row"].values
    )

    return xds


def is_nested_ms(attrs: Dict) -> bool:
    ctds_attrs = attrs["other"]["msv2"]["ctds_attrs"]
    return (
        "MS_VERSION" in ctds_attrs
        and "column_descriptions" in ctds_attrs
        and all(
            col in ctds_attrs["column_descriptions"]
            for col in (
                "UVW",
                "ANTENNA1",
                "ANTENNA2",
                "FEED1",
                "FEED2",
                "OBSERVATION_ID",
            )
        )
    )


def read_generic_table(
    inpath: str,
    tname: str,
    timecols: Union[List[str], None] = None,
    ignore=None,
    rename_ids: Dict[str, str] = None,
) -> xr.Dataset:
    """read generic casacore (sub)table into memory resident xds. This reads the table
    and and loads the data.

    Parameters
    ----------
    :param inpath: path to the MS or directory containing the table
    :param tname: (sub)table name, for example 'SOURCE' for myms.ms/SOURCE

    :param timecols: column names to convert to numpy datetime format.
    leaves times as their original casacore format.
    :param ignore: list of column names to ignore and not try to read.
    :rename_ids: dict with dimension renaming mapping

    :return: table loaded as XArray dataset
    """
    if timecols is None:
        timecols = []
    if ignore is None:
        ignore = []

    infile = Path(inpath, tname)
    infile = str(infile.expanduser())
    if not os.path.isdir(infile):
        raise ValueError(
            f"invalid input filename to read_generic_table: {infile} table {tname}"
        )

    cc_attrs = extract_table_attributes(infile)
    attrs: Dict[str, Any] = {"other": {"msv2": {"ctds_attrs": cc_attrs}}}
    if is_nested_ms(attrs):
        logging.warning(
            f"Skipping subtable that looks like a MeasurementSet main table: {inpath} {tname}"
        )
        return xr.Dataset()

    with open_table_ro(infile) as tb_tool:
        if tb_tool.nrows() == 0:
            logging.debug(f"table is empty: {inpath} {tname}")
            return xr.Dataset(attrs=attrs)

        colnames = tb_tool.colnames()
        mcoords, mvars = read_generic_cols(infile, tb_tool, timecols, ignore)

    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    xds = xr.Dataset(mvars, coords=mcoords)

    dim_prefix = "dim"
    dims = ["row"] + [f"{dim_prefix}_{i}" for i in range(1, 20)]
    xds = xds.rename(dict([(dv, dims[di]) for di, dv in enumerate(xds.dims)]))
    if rename_ids:
        rename_ids = {k: v for k, v in rename_ids.items() if k in xds.dims}
    xds = xds.rename_dims(rename_ids)

    attrs["other"]["msv2"]["bad_cols"] = list(
        np.setdiff1d(
            [dv for dv in colnames],
            [dv for dv in list(xds.data_vars) + list(xds.coords)],
        )
    )

    if tname in ["DOPPLER", "FREQ_OFFSET", "SOURCE", "SYSCAL", "WEATHER"]:
        xds = redimension_ms_subtable(xds, tname)

    if is_ephem_subtable(tname):
        xds = add_ephemeris_vars(tname, xds)
        xds = redimension_ms_subtable(xds, "EPHEMERIDES")

    xds = xds.assign_attrs(attrs)

    return xds


def read_generic_cols(
    infile: str, tb_tool, timecols, ignore, dim_prefix: str = "dim"
) -> Tuple[Dict[str, xr.Dataset], Dict[str, xr.Dataset]]:
    """Reads data for each MS column (loading the data in memory) into
    Xarray datasets

    :param infile: path name of the MS
    :param tb_tool: table to red the columns
    :param timecols: columns names to convert to datetime format
    :param ignore: list of column names to skip and not try to read.

    :return: dict of coordinates and dict of data vars.
    """

    # Almost sure that when TIME is present (in a standard MS subt) it
    # is part of the key. But what about non-std subtables, ASDM subts?
    subts_with_time_key = (
        "FEED",
        "FLAG_CMD",
        "FREQ_OFFSET",
        "HISTORY",
        "SOURCE",
        "SYSCAL",
        "WEATHER",
    )
    # Optional cols known to sometimes have inconsistent values
    known_misbehaving_cols = ["ASSOC_NATURE"]

    colnames = tb_tool.colnames()
    col_cells = {
        col: tb_tool.getcell(col, 0)
        for col in colnames
        if (col not in ignore) and (tb_tool.iscelldefined(col, 0))
    }

    trows = tb_tool.row(ignore, exclude=True)[:]

    # Produce coords and data vars from MS columns
    mcoords, mvars = {}, {}
    for col in col_cells.keys():
        if tb_tool.coldatatype(col) == "record":
            continue  # not supported

        try:
            data = np.stack(
                [row[col] for row in trows]
            )  # .astype(col_cells[col].dtype)
            if isinstance(trows[0][col], dict):
                data = np.stack(
                    [
                        row[col]["array"].reshape(row[col]["shape"])
                        if len(row[col]["array"]) > 0
                        else np.array([""])
                        for row in trows
                    ]
                )
        except Exception:
            # sometimes the cols are variable, so we need to standardize to the largest sizes
            if len(np.unique([isinstance(row[col], dict) for row in trows])) > 1:
                continue  # can't deal with this case
            mshape = np.array(max([np.array(row[col]).shape for row in trows]))
            try:
                pad_nan = get_pad_nan(col_cells[col])
                data = np.stack(
                    [
                        np.pad(
                            row[col]
                            if len(row[col]) > 0
                            else np.array(row[col]).reshape(np.arange(len(mshape)) * 0),
                            [(0, ss) for ss in mshape - np.array(row[col]).shape],
                            "constant",
                            constant_values=pad_nan,
                        )
                        for row in trows
                    ]
                )
            except Exception as exc:
                level = logging.WARNING
                if col in known_misbehaving_cols:
                    level = logging.DEBUG
                logging.log(
                    level, f"{infile}: failed to read data for column {col}: {exc}"
                )
                data = []

        if len(data) == 0:
            continue
        if col in timecols:
            if col == "MJD":
                data = convert_mjd_time(data)
            else:
                try:
                    data = convert_casacore_time(data)
                except pd.errors.OutOfBoundsDatetime as exc:
                    if infile.endswith("WEATHER"):
                        logging.error(
                            f"Exception when converting WEATHER/TIME: {exc}. TIME data: {data}"
                        )
                    else:
                        raise
        # should also probably add INTERVAL not only TIME
        if col.endswith("_ID") or (
            infile.endswith(subts_with_time_key) and col == "TIME"
        ):
            mcoords[col.lower()] = xr.DataArray(
                data,
                dims=[
                    f"{dim_prefix}_{di}_{ds}"
                    for di, ds in enumerate(np.array(data).shape)
                ],
            )
        else:
            mvars[col.lower()] = xr.DataArray(
                data,
                dims=[
                    f"{dim_prefix}_{di}_{ds}"
                    for di, ds in enumerate(np.array(data).shape)
                ],
            )

    return mcoords, mvars


def read_flat_col_chunk(infile, col, cshape, ridxs, cstart, pstart) -> np.ndarray:
    """
    Extract data chunk for each table col, this is fed to dask.delayed
    """

    with open_table_ro(infile) as tb_tool:
        rgrps = [
            (rr[0], rr[-1])
            for rr in np.split(ridxs, np.where(np.diff(ridxs) > 1)[0] + 1)
        ]
        # try:
        if (len(cshape) == 1) or (col == "UVW"):  # all the scalars and UVW
            data = np.concatenate(
                [tb_tool.getcol(col, rr[0], rr[1] - rr[0] + 1) for rr in rgrps], axis=0
            )
        elif len(cshape) == 2:  # WEIGHT, SIGMA
            data = np.concatenate(
                [
                    tb_tool.getcolslice(
                        col,
                        pstart,
                        pstart + cshape[1] - 1,
                        [],
                        rr[0],
                        rr[1] - rr[0] + 1,
                    )
                    for rr in rgrps
                ],
                axis=0,
            )
        elif len(cshape) == 3:  # DATA and FLAG
            data = np.concatenate(
                [
                    tb_tool.getcolslice(
                        col,
                        (cstart, pstart),
                        (cstart + cshape[1] - 1, pstart + cshape[2] - 1),
                        [],
                        rr[0],
                        rr[1] - rr[0] + 1,
                    )
                    for rr in rgrps
                ],
                axis=0,
            )
        # except:
        #    print('ERROR reading chunk: ', col, cshape, cstart, pstart)

    return data


def read_col_chunk(
    infile: str,
    ts_taql: str,
    col: str,
    cshape: Tuple[int],
    tidxs: np.ndarray,
    bidxs: np.ndarray,
    didxs: np.ndarray,
    d1: Tuple[int, int],
    d2: Tuple[int, int],
) -> np.ndarray:
    """
    Function to perform delayed reads from table columns.
    """
    # TODO: consider calling load_col_chunk() from inside the withs
    # for read_delayed_pointing_table and read_expanded_main_table
    with open_table_ro(infile) as mtable:
        with open_query(mtable, ts_taql) as query:
            if (len(cshape) == 2) or (col == "UVW"):  # all the scalars and UVW
                data = np.array(query.getcol(col, 0, -1))
            elif len(cshape) == 3:  # WEIGHT, SIGMA
                data = query.getcolslice(col, d1[0], d1[1], [], 0, -1)
            elif len(cshape) == 4:  # DATA and FLAG
                data = query.getcolslice(col, (d1[0], d2[0]), (d1[1], d2[1]), [], 0, -1)

    # full data is the maximum of the data shape and chunk shape dimensions
    fulldata = np.full(cshape, np.nan, dtype=data.dtype)
    if len(didxs) > 0:
        fulldata[tidxs[didxs], bidxs[didxs]] = data[didxs]

    return fulldata



import numpy as np

# Changed column_descriptions['UVW']['MEASINFO']['Ref'] = ITRF -> J2000 from deafult assigned by CASA simulator. Follows convention in fixvis.
# column_descriptions['UVW']['shape']
main_column_description = {
    "UVW": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 5,
        "maxlen": 0,
        "comment": "Vector with uvw coordinates (in meters)",
        "ndim": 1,
        "shape": np.array([3]),
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "uvw", "Ref": "ITRF"},
        },
    },
    "FLAG": {
        "valueType": "boolean",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "FlagColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "The data flags, array of bools with same shape as data",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "FLAG_CATEGORY": {
        "valueType": "boolean",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "flagHyperColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "The flag category, NUM_CAT flags for each datum",
        "ndim": 3,
        "_c_order": True,
        "keywords": {"CATEGORY": []},
    },
    "WEIGHT": {
        "valueType": "float",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "WeightColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "Weight for each polarization spectrum",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "SIGMA": {
        "valueType": "float",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "SigmaColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "Estimated rms noise for channel with unity bandpass response",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "ANTENNA1": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "SSM",
        "option": 0,
        "maxlen": 0,
        "comment": "ID of first antenna in interferometer",
        "keywords": {},
    },
    "ANTENNA2": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "SSM",
        "option": 0,
        "maxlen": 0,
        "comment": "ID of second antenna in interferometer",
        "keywords": {},
    },
    "ARRAY_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "ID of array or subarray",
        "keywords": {},
    },
    "DATA_DESC_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "The data description table index",
        "keywords": {},
    },
    "EXPOSURE": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "The effective integration time",
        "keywords": {"QuantumUnits": ["s"]},
    },
    "FEED1": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "The feed index for ANTENNA1",
        "keywords": {},
    },
    "FEED2": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "The feed index for ANTENNA2",
        "keywords": {},
    },
    "FIELD_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Unique id for this pointing",
        "keywords": {},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Row flag - flag all data in this row if True",
        "keywords": {},
    },
    "INTERVAL": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "The sampling interval",
        "keywords": {"QuantumUnits": ["s"]},
    },
    "OBSERVATION_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "ID for this observation, index in OBSERVATION table",
        "keywords": {},
    },
    "PROCESSOR_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Id for backend processor, index in PROCESSOR table",
        "keywords": {},
    },
    "SCAN_NUMBER": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Sequential scan number from on-line system",
        "keywords": {},
    },
    "STATE_ID": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "ID for this observing state",
        "keywords": {},
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Modified Julian Day",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "TIME_CENTROID": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ismdata",
        "option": 0,
        "maxlen": 0,
        "comment": "Modified Julian Day",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "DATA": {
        "valueType": "complex",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "dataHyperColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "The data column",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "MODEL_DATA": {
        "valueType": "complex",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "ModelDataColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "The model data column",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "CORRECTED_DATA": {
        "valueType": "complex",
        "dataManagerType": "TiledShapeStMan",
        "dataManagerGroup": "CorrectedDataColumn",
        "option": 0,
        "maxlen": 0,
        "comment": "The corrected data column",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "ROWID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "ROWID column",
        "keywords": {},
    },
}

# spectral_window_column_description['CHAN_FREQ']['keywords']['MEASINFO']['TabRefCodes']
# spectral_window_column_description['REF_FREQUENCY']['keywords']['MEASINFO']['TabRefCodes']
spectral_window_column_description = {
    "MEAS_FREQ_REF": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Frequency Measure reference",
        "keywords": {},
    },
    "CHAN_FREQ": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Center frequencies for each channel in the data matrix",
        "ndim": 1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["Hz"],
            "MEASINFO": {
                "type": "frequency",
                "VarRefCol": "MEAS_FREQ_REF",
                "TabRefTypes": [
                    "REST",
                    "LSRK",
                    "LSRD",
                    "BARY",
                    "GEO",
                    "TOPO",
                    "GALACTO",
                    "LGROUP",
                    "CMB",
                    "Undefined",
                ],
                "TabRefCodes": np.array(
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 64], dtype=np.uint32
                ),
            },
        },
    },
    "REF_FREQUENCY": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The reference frequency",
        "keywords": {
            "QuantumUnits": ["Hz"],
            "MEASINFO": {
                "type": "frequency",
                "VarRefCol": "MEAS_FREQ_REF",
                "TabRefTypes": [
                    "REST",
                    "LSRK",
                    "LSRD",
                    "BARY",
                    "GEO",
                    "TOPO",
                    "GALACTO",
                    "LGROUP",
                    "CMB",
                    "Undefined",
                ],
                "TabRefCodes": np.array(
                    [0, 1, 2, 3, 4, 5, 6, 7, 8, 64], dtype=np.uint32
                ),
            },
        },
    },
    "CHAN_WIDTH": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Channel width for each channel",
        "ndim": 1,
        "_c_order": True,
        "keywords": {"QuantumUnits": ["Hz"]},
    },
    "EFFECTIVE_BW": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Effective noise bandwidth of each channel",
        "ndim": 1,
        "_c_order": True,
        "keywords": {"QuantumUnits": ["Hz"]},
    },
    "RESOLUTION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The effective noise bandwidth for each channel",
        "ndim": 1,
        "_c_order": True,
        "keywords": {"QuantumUnits": ["Hz"]},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Row flag",
        "keywords": {},
    },
    "FREQ_GROUP": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Frequency group",
        "keywords": {},
    },
    "FREQ_GROUP_NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Frequency group name",
        "keywords": {},
    },
    "IF_CONV_CHAIN": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The IF conversion chain number",
        "keywords": {},
    },
    "NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Spectral window name",
        "keywords": {},
    },
    "NET_SIDEBAND": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Net sideband",
        "keywords": {},
    },
    "NUM_CHAN": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Number of spectral channels",
        "keywords": {},
    },
    "TOTAL_BANDWIDTH": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The total bandwidth for this window",
        "keywords": {"QuantumUnits": ["Hz"]},
    },
}


polarization_column_description = {
    "CORR_TYPE": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The polarization type for each correlation product, as a Stokes enum.",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "CORR_PRODUCT": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Indices describing receptors of feed going into correlation",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Row flag",
        "keywords": {},
    },
    "NUM_CORR": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Number of correlation products",
        "keywords": {},
    },
}


data_description_column_description = {
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Flag this row",
        "keywords": {},
    },
    "POLARIZATION_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Pointer to polarization table",
        "keywords": {},
    },
    "SPECTRAL_WINDOW_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Pointer to spectralwindow table",
        "keywords": {},
    },
}

# antenna_column_description['OFFSET']['shape']
# antenna_column_description['POSITION']['shape']
antenna_column_description = {
    "OFFSET": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 5,
        "maxlen": 0,
        "comment": "Axes offset of mount to FEED REFERENCE point",
        "ndim": 1,
        "shape": np.array([3]),
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "position", "Ref": "ITRF"},
        },
    },
    "POSITION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 5,
        "maxlen": 0,
        "comment": "Antenna X,Y,Z phase reference position",
        "ndim": 1,
        "shape": np.array([3]),
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "position", "Ref": "ITRF"},
        },
    },
    "TYPE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Antenna type (e.g. SPACE-BASED)",
        "keywords": {},
    },
    "DISH_DIAMETER": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Physical diameter of dish",
        "keywords": {"QuantumUnits": ["m"]},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Flag for this row",
        "keywords": {},
    },
    "MOUNT": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Mount type e.g. alt-az, equatorial, etc.",
        "keywords": {},
    },
    "NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Antenna name, e.g. VLA22, CA03",
        "keywords": {},
    },
    "STATION": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Station (antenna pad) name",
        "keywords": {},
    },
}

# feed_column_description['POSITION']['shape']
feed_column_description = {
    "POSITION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 5,
        "maxlen": 0,
        "comment": "Position of feed relative to feed reference position",
        "ndim": 1,
        "shape": np.array([3]),
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "position", "Ref": "ITRF"},
        },
    },
    "BEAM_OFFSET": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Beam position offset (on sky but in antennareference frame)",
        "ndim": 2,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "POLARIZATION_TYPE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Type of polarization to which a given RECEPTOR responds",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "POL_RESPONSE": {
        "valueType": "complex",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "D-matrix i.e. leakage between two receptors",
        "ndim": 2,
        "_c_order": True,
        "keywords": {},
    },
    "RECEPTOR_ANGLE": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "The reference angle for polarization",
        "ndim": 1,
        "_c_order": True,
        "keywords": {"QuantumUnits": ["rad"]},
    },
    "ANTENNA_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "ID of antenna in this array",
        "keywords": {},
    },
    "BEAM_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Id for BEAM model",
        "keywords": {},
    },
    "FEED_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Feed id",
        "keywords": {},
    },
    "INTERVAL": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Interval for which this set of parameters is accurate",
        "keywords": {"QuantumUnits": ["s"]},
    },
    "NUM_RECEPTORS": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Number of receptors on this feed (probably 1 or 2)",
        "keywords": {},
    },
    "SPECTRAL_WINDOW_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "ID for this spectral window setup",
        "keywords": {},
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Midpoint of time for which this set of parameters is accurate",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
}


field_column_description = {
    "DELAY_DIR": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Direction of delay center (e.g. RA, DEC)as polynomial in time.",
        "ndim": 2,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "PHASE_DIR": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Direction of phase center (e.g. RA, DEC).",
        "ndim": 2,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "REFERENCE_DIR": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Direction of REFERENCE center (e.g. RA, DEC).as polynomial in time.",
        "ndim": 2,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "CODE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Special characteristics of field, e.g. Bandpass calibrator",
        "keywords": {},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Row Flag",
        "keywords": {},
    },
    "NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Name of this field",
        "keywords": {},
    },
    "NUM_POLY": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Polynomial order of _DIR columns",
        "keywords": {},
    },
    "SOURCE_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Source id",
        "keywords": {},
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Time origin for direction and rate",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
}


history_column_description = {
    "APP_PARAMS": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Application parameters",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "CLI_COMMAND": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "CLI command sequence",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "APPLICATION": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Application name",
        "keywords": {},
    },
    "MESSAGE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Log message",
        "keywords": {},
    },
    "OBJECT_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Originating ObjectID",
        "keywords": {},
    },
    "OBSERVATION_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Observation id (index in OBSERVATION table)",
        "keywords": {},
    },
    "ORIGIN": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "(Source code) origin from which message originated",
        "keywords": {},
    },
    "PRIORITY": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Message priority",
        "keywords": {},
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Timestamp of message",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
}

# observation_column_description['TIME_RANGE']['shape']
observation_column_description = {
    "TIME_RANGE": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 5,
        "maxlen": 0,
        "comment": "Start and end of observation",
        "ndim": 1,
        "shape": np.array([2]),
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "LOG": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Observing log",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "SCHEDULE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Observing schedule",
        "ndim": 1,
        "_c_order": True,
        "keywords": {},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Row flag",
        "keywords": {},
    },
    "OBSERVER": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Name of observer(s)",
        "keywords": {},
    },
    "PROJECT": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Project identification string",
        "keywords": {},
    },
    "RELEASE_DATE": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Release date when data becomes public",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "SCHEDULE_TYPE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Observing schedule type",
        "keywords": {},
    },
    "TELESCOPE_NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Telescope Name (e.g. WSRT, VLBA)",
        "keywords": {},
    },
}


pointing_column_description = {
    "DIRECTION": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Antenna pointing direction as polynomial in time",
        "ndim": 2,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "ANTENNA_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "SSMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Antenna Id",
        "keywords": {},
    },
    "INTERVAL": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Time interval",
        "keywords": {"QuantumUnits": ["s"]},
    },
    "NAME": {
        "valueType": "string",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Pointing position name",
        "keywords": {},
    },
    "NUM_POLY": {
        "valueType": "int",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Series order",
        "keywords": {},
    },
    "TARGET": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "target direction as polynomial in time",
        "ndim": -1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Time interval midpoint",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "TIME_ORIGIN": {
        "valueType": "double",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Time origin for direction",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "TRACKING": {
        "valueType": "boolean",
        "dataManagerType": "IncrementalStMan",
        "dataManagerGroup": "ISMPointing",
        "option": 0,
        "maxlen": 0,
        "comment": "Tracking flag - True if on position",
        "keywords": {},
    },
}

source_column_description = {
    "CALIBRATION_GROUP": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Number of grouping for calibration purpose.",
        "keywords": {},
    },
    "CODE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Special characteristics of source, e.g. Bandpass calibrator",
        "keywords": {},
    },
    "DIRECTION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Direction (e.g. RA, DEC).",
        "ndim": -1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["rad", "rad"],
            "MEASINFO": {"type": "direction", "Ref": "J2000"},
        },
    },
    "INTERVAL": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Interval of time for which this set of parameters is accurate",
        "keywords": {"QuantumUnits": ["s"]},
    },
    "NAME": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Name of source as given during observations",
        "keywords": {},
    },
    "NUM_LINES": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Number of spectral lines",
        "keywords": {},
    },
    "PROPER_MOTION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Proper motion",
        "ndim": -1,
        "_c_order": True,
        "keywords": {"QuantumUnits": ["rad/s"]},
    },
    "SOURCE_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Source id",
        "keywords": {},
    },
    "SPECTRAL_WINDOW_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "ID for this spectral window setup",
        "keywords": {},
    },
    "TIME": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Midpoint of time for which this set of parameters is accurate.",
        "keywords": {
            "QuantumUnits": ["s"],
            "MEASINFO": {"type": "epoch", "Ref": "UTC"},
        },
    },
    "POSITION": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Position (e.g. for solar system objects",
        "ndim": -1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m", "m", "m"],
            "MEASINFO": {"type": "position", "Ref": "ITRF"},
        },
    },
    "PULSAR_ID": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Pulsar Id, pointer to pulsar table",
        "keywords": {},
    },
    "REST_FREQUENCY": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Line rest frequency",
        "ndim": -1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["Hz"],
            "MEASINFO": {"type": "frequency", "Ref": "LSRK"},
        },
    },
    "SYSVEL": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Systemic velocity at reference",
        "ndim": -1,
        "_c_order": True,
        "keywords": {
            "QuantumUnits": ["m/s"],
            "MEASINFO": {"type": "radialvelocity", "Ref": "LSRK"},
        },
    },
    "TRANSITION": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Line Transition name",
        "ndim": -1,
        "_c_order": True,
        "keywords": {},
    },
}


state_column_description = {
    "CAL": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Noise calibration temperature",
        "keywords": {"QuantumUnits": ["K"]},
    },
    "FLAG_ROW": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Row flag",
        "keywords": {},
    },
    "LOAD": {
        "valueType": "double",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Load temperature",
        "keywords": {"QuantumUnits": ["K"]},
    },
    "OBS_MODE": {
        "valueType": "string",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Observing mode, e.g., OFF_SPECTRUM",
        "keywords": {},
    },
    "REF": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "True for a reference observation",
        "keywords": {},
    },
    "SIG": {
        "valueType": "boolean",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "True for a source observation",
        "keywords": {},
    },
    "SUB_SCAN": {
        "valueType": "int",
        "dataManagerType": "StandardStMan",
        "dataManagerGroup": "StandardStMan",
        "option": 0,
        "maxlen": 0,
        "comment": "Sub scan number, relative to scan number",
        "keywords": {},
    },
}



import logging
from pathlib import Path
from typing import Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from casacore import tables

from ._tables.table_query import open_table_ro, open_query


def make_partition_ids_by_ddi_scan(
    infile: str, do_subscans: bool
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Produces arrays of per-partition ddi, scan, state_id, for when
    using partiion schemes 'scan' or 'scan/subscan', that is
    partitioning by some variant of (ddi, scan, subscan(state_id))

    :param infile: Path to MS
    :param do_subscans: also partitioning by subscan, not only scan
    :return: arrays with indices that define every partition
    """
    try:
        cctable = None
        taql_distinct_states = None
        cctable = tables.table(
            infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
        )
        if do_subscans:
            taql_distinct_states = (
                "select DISTINCT SCAN_NUMBER, STATE_ID, DATA_DESC_ID from $cctable"
            )
        else:
            taql_distinct_states = (
                "select DISTINCT SCAN_NUMBER, DATA_DESC_ID from $cctable"
            )
        with open_query(cctable, taql_distinct_states) as query_states:
            logging.debug(
                f"Got query, nrows: {query_states.nrows()}, query: {query_states}"
            )
            scan_number = query_states.getcol("SCAN_NUMBER")
            logging.debug(
                f"Got col SCAN_NUMBER (len: {len(scan_number)}): {scan_number}"
            )
            if do_subscans:
                state_id = query_states.getcol("STATE_ID")
                data_desc_id = [None] * len(scan_number)
            else:
                state_id = [None] * len(scan_number)
                logging.debug(f"Got col STATE_ID (len: {len(state_id)}): {state_id}")
                data_desc_id = query_states.getcol("DATA_DESC_ID")

        logging.debug(
            f"Got col DATA_DESC_ID (len: {len(data_desc_id)}): {data_desc_id}"
        )
        logging.debug(
            f"Len of DISTINCT SCAN_NUMBER,etc.: {len(scan_number)}. Will generate that number of partitions"
        )
    finally:
        if cctable:
            cctable.close()

    return data_desc_id, scan_number, state_id


def partition_when_empty_state(
    infile: str,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Generate fallback partition ids when trying to partition by
    'intent' but the STATE table is empty.

    Some MSs have no STATE rows and in the main table STATE_ID==-1
    (that is not a valid MSv2 but it happens).

    :param infile: Path to the MS
    :return: same as make_partition_ids_by_ddi_intent but with
    effectively only ddi indices and other indices set to None ("any
    IDs found")
    """
    try:
        main_table = None

        main_table = tables.table(
            infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
        )
        taql_ddis = "select DISTINCT DATA_DESC_ID from $main_table"
        with open_query(main_table, taql_ddis) as query_per_intent:
            # Will take whatever scans given the STATE_IDs and DDIs
            # scan_number = query_per_intent.getcol("SCAN_NUMBER")
            distinct_ddis = query_per_intent.getcol("DATA_DESC_ID")

        logging.debug(
            f"Producing {len(distinct_ddis)} partitions for ddis: {distinct_ddis}"
        )
        nparts = len(distinct_ddis)

    finally:
        if main_table:
            main_table.close()

    return distinct_ddis, [None] * nparts, [None] * nparts, [""] * nparts


def find_distinct_obs_mode(
    infile: str, state_table: tables.table
) -> Union[List[str], None]:
    """Produce a list of unique "scan/subscan" intents.

    :param infile: Path to the MS
    :param state_table: casacore table object to read from
    :return: List of unique "scan/subscan" intents as given in the
    OBS_MODE column of the STATE subtable. None if the STATE subtable
    is empty or there is a problem reading it

    """
    taql_distinct_intents = "select DISTINCT OBS_MODE from $state_table"
    with open_query(state_table, taql_distinct_intents) as query_intents:
        if query_intents.nrows() == 0:
            logging.warning(
                "STATE subtable has no data. Cannot partition by scan/subscan intent"
            )
            return None

        distinct_obs_mode = query_intents.getcol("OBS_MODE")
        logging.debug(
            f"  Query for distinct OBS_MODE len: {len(distinct_obs_mode)}, values: {distinct_obs_mode}"
        )
        return distinct_obs_mode


def filter_intents_per_ddi(
    ddis: List[int], substr: str, intents: str, spw_name_by_ddi: Dict[int, str]
) -> List[str]:
    """For a given pair of:
    - substring (say 'WVR') associated with a type of intent we want to differentiate
    - intents string (multiple comma-separated scan/subscan intents)
    => do: for every DDI passed in the list of ddis, either keep only the
    intents that have that substring (if there are any) or drop them, depending on
    whether that substring is present in the SPW name. This is to filter in only
    the intents that really apply to every DDI/SPW.

    :param ddis: list of ddis for which the intents have to be filtered
    :param substr: substring to filter by
    :param intents: string with a comma-separated list of individual
    scan/subscan intent strings (like scan/subscan intents as stored
    in the MS STATE/OBS_MODE
    :param spw_name_by_ddi: SPW names by DDI ID (row index) key
    :return: list where the intents related to 'substr' have been filtered in our out

    """
    present = substr in intents
    # Nothing to effectively filter, full cs-list of intents apply to all DDIs
    if not present:
        return [intents] * len(ddis)

    every_intent = intents.split(",")
    filtered_intents = []
    for ddi in ddis:
        spw_name = spw_name_by_ddi[ddi]
        if not spw_name:
            # we cannot say / cannot filter
            filtered_intents.append(intents)
            continue

        # A not-xor to select/deselect (or keep-only/drop) the intents that apply
        # to this DDI
        ddi_intents = [
            intnt for intnt in every_intent if (substr in intnt) == (substr in spw_name)
        ]
        ddi_intents = ",".join(ddi_intents)
        filtered_intents.append(ddi_intents)

    return filtered_intents


def make_ddi_state_intent_lists(
    main_tbl: tables.table,
    state_tbl: tables.table,
    distinct_obs_mode: np.ndarray,
    spw_name_by_ddi: Dict[int, str],
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Produce arrays of (ddi indices, state indices, intent string)
    for every distinct intent string, where every item represents one
    partition of the main table

    As the obs_mode strings have concatenated intent strings from all
    the scan and subscan intents, this function has started
    implementing some simple heuristics to remove the intent items
    that are not related to the respective DDIs (for example WVR
    intent is the only kept when the DDI/SPW has WVR in its name). See
    call to filter_intents_per_ddi()

    :param main_tbl: main MS table openend as a casacore.tables.table
    :state_tbl: STATE subtable openend as a casacore.tables.table
    :param distinct_obs_mode: list of unique/distinct OBS_MODE strings from the STATE table
    :return: arrays of (ddi indices, state indices, intent string)

    """
    data_desc_id, state_id_partitions, intent_names = [], [], []
    for intent in distinct_obs_mode:
        where_intent = f"where OBS_MODE = '{intent}'"
        taql_states = f"select ROWID() as ROWS from $state_tbl {where_intent}"
        with open_query(state_tbl, taql_states) as query_states:
            state_ids_for_intent = query_states.getcol("ROWS")

        state_ids = " OR STATE_ID = ".join(np.char.mod("%d", state_ids_for_intent))
        taql_partition = (
            f"select DISTINCT DATA_DESC_ID from $main_tbl where STATE_ID = {state_ids}"
        )
        with open_query(main_tbl, taql_partition) as query_ddi_intent:
            # No data for these STATE_IDs
            if query_ddi_intent.nrows() == 0:
                continue

            # Will implicitly take whatever scans given the STATE_IDs
            # and DDIs scan_number. Not needed:
            # scan_number = query_ddi_intent.getcol("SCAN_NUMBER")
            ddis = query_ddi_intent.getcol("DATA_DESC_ID")

            data_desc_id.extend(ddis)
            state_id_partitions.extend([state_ids_for_intent] * len(ddis))

            # Try to select/exclude confusing or mixed intent names such as 'WVR#*'
            intents_ddi = filter_intents_per_ddi(ddis, "WVR", intent, spw_name_by_ddi)
            intent_names.extend(intents_ddi)

    logging.debug(
        f"Produced data_desc_id: {data_desc_id},\n state_id_partitions: {state_id_partitions}"
    )
    return data_desc_id, state_id_partitions, intent_names


def make_partition_ids_by_ddi_intent(
    infile: str, spw_names: xr.DataArray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Produces arrays of per-partition ddi, scan, state_id, for when
    using the partiion scheme 'intents' (ddi, scan, subscans(state_ids))

    :param infile:
    :return: arrays with indices that define every partition
    """
    # TODO: could explore other TAQL alternatives, like
    # select ... from ::STATE where OBS_MODE = ...
    #
    # This will work only if intents are already alphabetically sorted (grouped),
    # won't work for alternating intents:
    # taql_intents = "select rowid() as ROWS from $state_tbl GROUPBY OBS_MODE "

    with open_table_ro(str(Path(infile, "STATE"))) as state_tbl:
        distinct_obs_mode = find_distinct_obs_mode(infile, state_tbl)

        if distinct_obs_mode is None:
            return partition_when_empty_state(infile)

        with open_table_ro(infile) as main_tbl:
            (
                data_desc_id,
                state_id_partitions,
                intent_names,
            ) = make_ddi_state_intent_lists(
                main_tbl, state_tbl, distinct_obs_mode, spw_names
            )

    # Take whatever scans given by the STATE_IDs and DDIs
    scan_number = [None] * len(state_id_partitions)

    return data_desc_id, scan_number, state_id_partitions, intent_names



import logging, os
from typing import Dict, List, Tuple, Union

import xarray as xr

from ._utils.cds import CASAVisSet
from ._ms.chunks import load_main_chunk
from ._ms.partitions import (
    finalize_partitions,
    read_ms_ddi_partitions,
    read_ms_scan_subscan_partitions,
)
from ._ms.subtables import read_ms_subtables
from ._utils.xds_helper import vis_xds_packager_cds

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)


def read_ms(
    infile: str,
    subtables: bool = True,
    asdm_subtables: bool = False,
    partition_scheme: str = "intent",
    chunks: Union[Tuple[int], List[int]] = None,
    expand: bool = False,
    **kwargs: str,
) -> CASAVisSet:
    """Read a MeasurementSet (MSv2 format) into a next generation CASA
    dataset (visibilities dataset as a set of Xarray datasets).

    The MS is partitioned into multiple sub- Xarray datasets (where the data variables are read as
    Dask delayed arrays).
    The MS is partitioned by DDI, which guarantees a fixed data shape per partition (in terms of channels
    and polarizations) and, subject to experimentation, by scan and subscan. This results in multiple
    partitions as xarray datasets (xds) contained within a main xds (mxds).

    :param infile: Input MS filename
    :param subtables: Also read and include subtables along with main table selection. Default False will
    omit subtables (faster)
    :param asdm_subtables: in addition to MeasurementSet subtables (if enabled), also read extension
    subtables named "ASDM_*"
    :param partition_scheme: (experimenting) Whether to partition sub-xds datasets by scan/subscan
    (in addition to DDI), or other alternative partitioning schemes. Accepted values: 'scan/subscan',
    'scan', 'ddi', 'intent'. Default: 'intent'
    :param chunks: Can be used to set a specific chunk shape (with a tuple of ints), or to control the
    optimization used for automatic chunking (with a list of ints). A tuple of ints in the form of (row,
    chan, pol) will use a fixed chunk shape. A list or numpy array of ints in the form of [idx1, etc]
    will trigger auto-chunking optimized for the given indices, with row=0, chan=1, pol=2. Default None
    uses auto-chunking with a best fit across all dimensions (probably sub-optimal for most cases).
    :param expand: (to be removed) Whether or not to return the original flat row structure of the MS (False)
    or expand the rows to time x baseline dimensions (True). Expanding the rows allows for easier indexing
    and parallelization across time and baseline dimensions, at the cost of some conversion time. Default
    False
    :param **kwargs: (to be removed?) Selection parameters from the standard way of making CASA MS
    selections. Supported keys are: spw, field, scan, baseline, time, scanintent, uvdist, polarization,
    array, observation.  Values are strings.

    :return: Main xarray dataset of datasets for this visibility dataset
    """

    infile = os.path.expanduser(infile)
    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to read_ms {infile}")

    # Several alternatives to experiment for now
    part_descr = {
        "intent": "scan/subscan intent + DDI",
        "ddi": "DDI",
        "scan": "scan + DDI",
        "scan/subscan": "scan + subscan + DDI",
    }

    if partition_scheme not in part_descr:
        raise ValueError(f"Invalid partition_scheme: {partition_scheme}")

    logging.info(
        f"Reading {infile} as MSv2 and applying partitioning by {part_descr[partition_scheme]}"
    )

    if partition_scheme == "ddi":
        logging.info(f"Reading {infile} as MSv2 and applying DDI partitioning")
        # get the indices of the ms selection (if any)
        # rowmap = ms_selection(infile, verbose=verbose, **kwargs)
        rowmap = None
        parts, subts, done_subts = read_ms_ddi_partitions(
            infile, expand, rowmap, chunks
        )
    else:
        parts, subts, done_subts = read_ms_scan_subscan_partitions(
            infile, partition_scheme, expand, chunks
        )

    if subtables:
        subts.update(read_ms_subtables(infile, done_subts, asdm_subtables))

    parts = finalize_partitions(parts, subts)

    # build the visibilities container (metainfo + partitions) to return
    cds = vis_xds_packager_cds(subts, parts, "read_ms")
    return cds


def load_vis_chunk(
    infile: str,
    block_des: Dict[str, slice],
    partition_key: Tuple[int, int, str],
) -> Dict[Tuple[int, int], xr.Dataset]:
    """Read a chunk of a MeasurementSet (MSv2 format) into an Xarray
    dataset, loading the data in memory.

    :param infile: Input MS filename
    :param block_des: specification of chunk to load

    :return: Xarray datasets with chunk of visibility data, one per DDI
    (spw_id, pol_setup_id pair)
    """
    infile = os.path.expanduser(infile)

    logging.info(f"Loading from {infile} as MSv2 a chunk of data into memory")

    if not os.path.isdir(infile):
        raise ValueError(f"invalid input filename to read_ms {infile}")

    orig_chunk_to_improve = load_main_chunk(infile, block_des)
    res = vis_xds_packager_cds(
        subtables={},
        partitions={partition_key: orig_chunk_to_improve},
        descr_add="load_vis_block",
    )
    return res



import os
import numpy as np
import json

from numcodecs.compat import (
    ensure_text,
    ensure_ndarray_like,
    ensure_bytes,
    ensure_contiguous_ndarray_like,
)


def pad_array_with_nans(input_array, output_shape, dtype):
    """
    Pad an integer array with NaN values to match the specified output shape.

    Parameters:
    - input_array: The input NumPy array to be padded.
    - output_shape: A tuple specifying the desired output shape (e.g., (rows, columns)).

    Returns:
    - A NumPy array with NaN padding to match the specified output shape.
    """
    # Get the input shape
    input_shape = input_array.shape

    # Calculate the padding dimensions
    padding_shape = tuple(max(0, o - i) for i, o in zip(input_shape, output_shape))

    # Create a new array filled with NaN values
    padded_array = np.empty(output_shape, dtype=dtype)
    padded_array[:] = np.nan

    # Copy the input array to the appropriate position within the padded array
    padded_array[: input_shape[0], : input_shape[1]] = input_array

    return padded_array


def write_binary_blob_to_disk(arr, file_path, compressor):
    """
    Compress a NumPy array using Blosc and save it to disk.

    Parameters:
    - arr: The NumPy array to compress and save.
    - file_path: The path to the output file where the compressed array will be saved.
    - compressor:

    Returns:
    - None
    """
    # Encode the NumPy array using the codec
    compressed_arr = compressor.encode(np.ascontiguousarray(arr))

    # Ensure the directory exists before saving the file
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    # Save the compressed array to disk
    with open(file_path, "wb") as file:
        file.write(compressed_arr)

    # print(f"Compressed array saved to {file_path}")


def read_binary_blob_from_disk(file_path, compressor, dtype=np.float64):
    """
    Read a compressed binary blob from disk and decode it using Blosc.

    Parameters:
    - file_path: The path to the compressed binary blob file.
    - compressor: The Blosc compressor to use (e.g., 'zstd', 'lz4', 'blosclz', etc.).

    Returns:
    - The decoded NumPy array.
    """
    # Check if the file exists
    if not os.path.exists(file_path):
        print(f"Error: File '{file_path}' not found.")
        return None

    # Read the compressed binary blob from disk
    with open(file_path, "rb") as file:
        compressed_arr = file.read()

    # Decode the compressed data using the Blosc compressor
    decoded_bytes = compressor.decode(compressed_arr)

    decoded_arr = np.frombuffer(decoded_bytes, dtype)  # Adjust dtype as needed

    return decoded_arr


def read_json_file(file_path):
    """
    Read a JSON file and return its contents as a Python dictionary.

    Parameters:
    - file_path: The path to the JSON file to be read.

    Returns:
    - A Python dictionary containing the JSON data.
    """
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
        return data
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return None
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON from '{file_path}': {e}")
        return None


class NumberEncoder(json.JSONEncoder):
    def default(self, o):
        # See json.JSONEncoder.default docstring for explanation
        # This is necessary to encode numpy dtype
        if isinstance(o, numbers.Integral):
            return int(o)
        if isinstance(o, numbers.Real):
            return float(o)
        return json.JSONEncoder.default(self, o)


def write_json_file(data, file_path):
    """
    Write a Python dictionary to a JSON file.

    Parameters:
    - data: The Python dictionary to be written to the JSON file.
    - file_path: The path to the JSON file to be created or overwritten.

    Returns:
    - None
    """
    with open(file_path, "w") as file:
        json.dump(
            data,
            file,
            indent=4,
            sort_keys=True,
            ensure_ascii=True,
            separators=(",", ": "),
            cls=NumberEncoder,
        )


def create_data_variable_meta_data_on_disk(
    zarr_group_name, data_variables_and_dims, xds_dims, parallel_coords, compressor
):
    zarr_meta = data_variables_and_dims

    for data_variable_key, dims_dtype_name in data_variables_and_dims.items():
        # print(data_variable_key, dims_dtype_name)

        dims = dims_dtype_name["dims"]
        dtype = dims_dtype_name["dtype"]
        data_variable_name = dims_dtype_name["name"]
        data_variable_path = os.path.join(zarr_group_name, data_variable_name)
        os.system("mkdir " + data_variable_path)
        # Create .zattrs
        zattrs = {
            "_ARRAY_DIMENSIONS": dims,
            # "coordinates": "time declination right_ascension"
        }

        shape = []
        chunks = []
        for d in dims:
            shape.append(xds_dims[d])
            if d in parallel_coords:
                chunks.append(len(parallel_coords[d]["data_chunks"][0]))
            else:
                chunks.append(xds_dims[d])

        # print(chunks,shape)
        write_json_file(zattrs, os.path.join(data_variable_path, ".zattrs"))

        # Create .zarray
        from zarr import n5

        compressor_config = n5.compressor_config_to_zarr(
            n5.compressor_config_to_n5(compressor.get_config())
        )

        if "f" in dtype:
            zarray = {
                "chunks": chunks,
                "compressor": compressor_config,
                "dtype": dtype,
                "fill_value": "NaN",
                "filters": None,
                "order": "C",
                "shape": shape,
                "zarr_format": 2,
            }

        else:
            zarray = {
                "chunks": chunks,
                "compressor": compressor_config,
                "dtype": dtype,
                "fill_value": None,
                "filters": None,
                "order": "C",
                "shape": shape,
                "zarr_format": 2,
            }

        zarr_meta[data_variable_key]["chunks"] = chunks
        zarr_meta[data_variable_key]["shape"] = shape

        write_json_file(zarray, os.path.join(data_variable_path, ".zarray"))
    return zarr_meta



import logging
from typing import Any, Dict, List, Tuple, Union

import dask, dask.array
import numpy as np
import xarray as xr

from casacore import tables

from .read import (
    read_flat_col_chunk,
    read_col_chunk,
    convert_casacore_time,
    extract_table_attributes,
    add_units_measures,
)

from .table_query import open_table_ro, open_query

rename_msv2_cols = {
    "antenna1": "antenna1_id",
    "antenna2": "antenna2_id",
    "feed1": "feed1_id",
    "feed2": "feed2_id",
    # optional cols:
    "weight_spectrum": "weight",
    "corrected_data": "vis_corrected",
    "data": "vis",
    "model_data": "vis_model",
    "float_data": "autocorr",
}


def rename_vars(mvars: Dict[str, xr.DataArray]) -> Dict[str, xr.DataArray]:
    """Apply rename rules. Also preserve ordering of data_vars

    Note: not using xr.DataArray.rename because we have optional
    column renames and rename complains if some of the names passed
    are not present in the dataset

    :param mvars: dictionary of data_vars to be used to create an xr.Dataset
    :return: similar dictionary after applying MSv2 => MSv3/ngCASA renaming rules
    """
    renamed = {
        rename_msv2_cols[name] if name in rename_msv2_cols else name: var
        for name, var in mvars.items()
    }

    return renamed


def redim_id_data_vars(mvars: Dict[str, xr.DataArray]) -> Dict[str, xr.DataArray]:
    """
    Changes:
    Several id data variables to drop its baseline dim
    The antenna id data vars:
     From MS (antenna1_id(time, baseline), antenna2_id(time,baseline)
     To cds (baseline_ant1_id(baseline), baseline_ant2_id(baseline)
    :param mvars: data variables being prepared for a partition xds
    :return: data variables with the ant id ones modified to cds type
    """
    # Vars to drop baseline dim
    var_names = [
        "array_id",
        "observation_id",
        "processor_id",
        "scan_number",
        "state_id",
    ]
    for vname in var_names:
        mvars[vname] = mvars[vname].sel(baseline=0, drop=True)

    for idx in ["1", "2"]:
        new_name = f"baseline_ant{idx}_id"
        mvars[new_name] = mvars.pop(f"antenna{idx}_id")
        mvars[new_name] = mvars[new_name].sel(time=0, drop=True)

    return mvars


def get_partition_ids(mtable: tables.table, taql_where: str) -> Dict:
    """Get some of the partition IDs that we have to retrieve from some
    of the top level ID/sorting cols of the main table of the MS.

    :param mtable: MS main table
    :param taql_where: where part that defines the partition in TaQL
    :return: ids of array, observation, and processor
    """

    taql_ids = f"select DISTINCT ARRAY_ID, OBSERVATION_ID, PROCESSOR_ID from $mtable {taql_where}"
    with open_query(mtable, taql_ids) as query:
        # array_id, observation_id, processor_id
        array_id = np.unique(query.getcol("ARRAY_ID"))
        obs_id = np.unique(query.getcol("OBSERVATION_ID"))
        proc_id = np.unique(query.getcol("PROCESSOR_ID"))
        check_vars = [
            (array_id, "array_id"),
            (obs_id, "observation_id"),
            (proc_id, "processor_id"),
        ]
        for var, var_name in check_vars:
            if len(var) != 1:
                logging.warning(
                    f"Did not get exactly one {var_name} (got {var} for this partition. TaQL: {taql_where}"
                )

    pids = {
        "array_id": list(array_id),
        "observation_id": list(obs_id),
        "processor_id": list(proc_id),
    }
    return pids


def read_expanded_main_table(
    infile: str,
    ddi: int = 0,
    scan_state: Union[Tuple[int, int], None] = None,
    ignore_msv2_cols: Union[list, None] = None,
    chunks: Tuple[int, ...] = (400, 200, 100, 2),
) -> Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]:
    """
    Reads one partition from the main table, all columns.
    This is the expanded version (time, baseline) dims.

    Chunk tuple: (time, baseline, freq, pol)
    """
    if ignore_msv2_cols is None:
        ignore_msv2_cols = []

    taql_where = f"where DATA_DESC_ID = {ddi}"
    if scan_state:
        # get partitions by scan/state
        scan, state = scan_state
        if type(state) == np.ndarray:
            state_ids_or = " OR STATE_ID = ".join(np.char.mod("%d", state))
            taql_where += f" AND (STATE_ID = {state_ids_or})"
        elif state:
            taql_where += f" AND SCAN_NUMBER = {scan} AND STATE_ID = {state}"
        elif scan:
            # scan can also be None, when partition_scheme='intent'
            # but the STATE table is empty!
            taql_where += f" AND SCAN_NUMBER = {scan}"

    with open_table_ro(infile) as mtable:
        # one partition, select just the specified ddi (+ scan/subscan)
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:
            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            xds, attrs = read_main_table_chunks(
                infile, tb_tool, taql_where, ignore_msv2_cols, chunks
            )
            part_ids = get_partition_ids(tb_tool, taql_where)

    return xds, part_ids, attrs


def read_main_table_chunks(
    infile: str,
    tb_tool: tables.table,
    taql_where: str,
    ignore_msv2_cols: Union[list, None] = None,
    chunks: Tuple[int, ...] = (400, 200, 100, 2),
) -> Tuple[xr.Dataset, Dict[str, Any]]:
    """
    Iterates through the time,baseline chunks and reads slices from
    all the data columns.
    """
    baselines = get_baselines(tb_tool)

    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]
    chan_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]

    utimes, tol = get_utimes_tol(tb_tool, taql_where)

    tvars = {}
    # loop over time chunks
    for tc in range(0, len(utimes), chunks[0]):
        times = (
            utimes[tc] - tol,
            utimes[min(len(utimes) - 1, tc + chunks[0] - 1)] + tol,
        )
        # chunk time length
        ctlen = min(len(utimes), tc + chunks[0]) - tc

        bvars = {}
        # loop over baseline chunks
        for bc in range(0, len(baselines), chunks[1]):
            blines = (
                baselines[bc],
                baselines[min(len(baselines) - 1, bc + chunks[1] - 1)],
            )
            cblen = min(len(baselines) - bc, chunks[1])

            # read the specified chunk of data
            # def read_chunk(infile, ddi, times, blines, chans, pols):
            ttql = f"TIME BETWEEN {times[0]} and {times[1]}"
            ants = (int(blines[0].split("_")[0]), int(blines[1].split("_")[0]))
            atql = f"ANTENNA1 BETWEEN {ants[0]} and {ants[1]}"
            ts_taql = f"select * from $mtable {taql_where} AND {ttql} AND {atql}"
            with open_query(None, ts_taql) as query_times_ants:
                tidxs = (
                    np.searchsorted(utimes, query_times_ants.getcol("TIME", 0, -1)) - tc
                )
                ts_ant1, ts_ant2 = (
                    query_times_ants.getcol("ANTENNA1", 0, -1),
                    query_times_ants.getcol("ANTENNA2", 0, -1),
                )

            ts_bases = [
                str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
                for ll in np.hstack([ts_ant1[:, None], ts_ant2[:, None]])
            ]
            bidxs = np.searchsorted(baselines, ts_bases) - bc

            # some antenna 2's will be out of bounds for this chunk, store rows that are in bounds
            didxs = np.where(
                (bidxs >= 0) & (bidxs < min(chunks[1], len(baselines) - bc))
            )[0]

            delayed_params = (infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs)

            read_all_cols_bvars(
                tb_tool, chunks, chan_cnt, ignore_msv2_cols, delayed_params, bvars
            )

        concat_bvars_update_tvars(bvars, tvars)

    dims = ["time", "baseline", "freq", "pol"]
    mvars = concat_tvars_to_mvars(dims, tvars, pol_cnt, chan_cnt)
    mcoords = {
        "time": xr.DataArray(convert_casacore_time(utimes), dims=["time"]),
        "baseline": xr.DataArray(np.arange(len(baselines)), dims=["baseline"]),
    }

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore_msv2_cols}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    return xds, attrs


def get_utimes_tol(mtable: tables.table, taql_where: str) -> Tuple[np.ndarray, float]:
    taql_utimes = f"select DISTINCT TIME from $mtable {taql_where}"
    with open_query(mtable, taql_utimes) as query_utimes:
        utimes = np.unique(query_utimes.getcol("TIME", 0, -1))
        # add a tol around the time ranges returned by taql
        if len(utimes) < 2:
            tol = 1e-5
        else:
            tol = np.diff(utimes).min() / 4

    return utimes, tol


def get_baselines(tb_tool: tables.table) -> np.ndarray:
    # main table uses time x (antenna1,antenna2)
    ant1, ant2 = tb_tool.getcol("ANTENNA1", 0, -1), tb_tool.getcol("ANTENNA2", 0, -1)
    baselines = np.array(
        [
            str(ll[0]).zfill(3) + "_" + str(ll[1]).zfill(3)
            for ll in np.unique(np.hstack([ant1[:, None], ant2[:, None]]), axis=0)
        ]
    )

    return baselines


def read_all_cols_bvars(
    tb_tool: tables.table,
    chunks: Tuple[int, ...],
    chan_cnt: int,
    ignore_msv2_cols,
    delayed_params: Tuple,
    bvars: Dict[str, xr.DataArray],
) -> None:
    """
    Loops over each column and create delayed dask arrays
    """

    col_names = tb_tool.colnames()
    for col in col_names:
        if (col in ignore_msv2_cols + ["TIME"]) or (not tb_tool.iscelldefined(col, 0)):
            continue
        if col not in bvars:
            bvars[col] = []

        cdata = tb_tool.getcol(col, 0, 1)[0]

        if len(cdata.shape) == 0:
            infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
            cshape = (ctlen, cblen)
            delayed_col = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
            delayed_array = dask.delayed(read_col_chunk)(*delayed_col, None, None)
            bvars[col] += [dask.array.from_delayed(delayed_array, cshape, cdata.dtype)]

        elif col == "UVW":
            infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
            cshape = (ctlen, cblen, 3)
            delayed_3 = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
            delayed_array = dask.delayed(read_col_chunk)(*delayed_3, None, None)
            bvars[col] += [dask.array.from_delayed(delayed_array, cshape, cdata.dtype)]

        elif len(cdata.shape) == 1:
            pol_list = []
            dd = 2 if cdata.shape == chan_cnt else 3
            for pc in range(0, cdata.shape[0], chunks[dd]):
                pols = (pc, min(cdata.shape[0], pc + chunks[dd]) - 1)
                infile, ts_taql, (ctlen, cblen), tidxs, bidxs, didxs = delayed_params
                cshape = (
                    ctlen,
                    cblen,
                ) + (pols[1] - pols[0] + 1,)
                delayed_cs = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
                delayed_array = dask.delayed(read_col_chunk)(*delayed_cs, pols, None)
                pol_list += [
                    dask.array.from_delayed(delayed_array, cshape, cdata.dtype)
                ]
            bvars[col] += [dask.array.concatenate(pol_list, axis=2)]

        elif len(cdata.shape) == 2:
            chan_list = []
            for cc in range(0, cdata.shape[0], chunks[2]):
                chans = (cc, min(cdata.shape[0], cc + chunks[2]) - 1)
                pol_list = []
                for pc in range(0, cdata.shape[1], chunks[3]):
                    pols = (pc, min(cdata.shape[1], pc + chunks[3]) - 1)
                    (
                        infile,
                        ts_taql,
                        (ctlen, cblen),
                        tidxs,
                        bidxs,
                        didxs,
                    ) = delayed_params
                    cshape = (
                        ctlen,
                        cblen,
                    ) + (chans[1] - chans[0] + 1, pols[1] - pols[0] + 1)
                    delayed_cs = infile, ts_taql, col, cshape, tidxs, bidxs, didxs
                    delayed_array = dask.delayed(read_col_chunk)(
                        *delayed_cs, chans, pols
                    )
                    pol_list += [
                        dask.array.from_delayed(delayed_array, cshape, cdata.dtype)
                    ]
                chan_list += [dask.array.concatenate(pol_list, axis=3)]
            bvars[col] += [dask.array.concatenate(chan_list, axis=2)]


def concat_bvars_update_tvars(
    bvars: Dict[str, xr.DataArray], tvars: Dict[str, xr.DataArray]
) -> None:
    """
    concats all the dask chunks from each baseline. This is intended to
    be called iteratively, for every time chunk iteration, once all the
    baseline chunks have been read.
    """
    for kk in bvars.keys():
        if len(bvars[kk]) == 0:
            continue
        if kk not in tvars:
            tvars[kk] = []
        tvars[kk] += [dask.array.concatenate(bvars[kk], axis=1)]


def concat_tvars_to_mvars(
    dims: List[str], tvars: Dict[str, xr.DataArray], pol_cnt: int, chan_cnt: int
) -> Dict[str, xr.DataArray]:
    """
    Concat into a single dask array all the dask arrays from each time
    chunk to make the final arrays of the xds.

    :param dims: dimension names
    :param tvars: variables as lists of dask arrays per time chunk
    :param pol_cnt: len of pol axis/dim
    :param chan_cnt: len of freq axis/dim (chan indices)

    :return: variables as concated dask arrays
    """

    mvars = {}
    for tvr in tvars.keys():
        data_var = tvr.lower()
        if tvr == "UVW":
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0),
                dims=dims[:2] + ["uvw_coords"],
            )
        elif len(tvars[tvr][0].shape) == 3 and (tvars[tvr][0].shape[-1] == pol_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0), dims=dims[:2] + ["pol"]
            )
        elif len(tvars[tvr][0].shape) == 3 and (tvars[tvr][0].shape[-1] == chan_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0), dims=dims[:2] + ["freq"]
            )
        else:
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(tvars[tvr], axis=0),
                dims=dims[: len(tvars[tvr][0].shape)],
            )

    return mvars


def read_flat_main_table(
    infile: str,
    ddi: Union[int, None] = None,
    scan_state: Union[Tuple[int, int], None] = None,
    rowidxs=None,
    ignore_msv2_cols: Union[List, None] = None,
    chunks: Tuple[int, ...] = (22000, 512, 2),
) -> Tuple[xr.Dataset, Dict[str, Any], Dict[str, Any]]:
    """
    Read main table using flat structure: no baseline dimension. Vis
    dimensions are: row, freq, pol
    (experimental, perhaps to be deprecated/removed). Works but some
    features may be missing and/or flaky.

    Chunk tuple: (row, freq, pol)
    """
    taql_where = f"where DATA_DESC_ID = {ddi}"
    if scan_state:
        # TODO: support additional intent/scan/subscan conditions if
        # we keep this read_flat functionality
        _scans, states = scan_state
        # get row indices relative to full main table
        if states:
            taql_where += (
                f" AND SCAN_NUMBER = {scan_state[0]} AND STATE_ID = {scan_state[1]}"
            )
        else:
            taql_where += f" AND SCAN_NUMBER = {scan_state[0]}"

    mtable = tables.table(
        infile, readonly=True, lockoptions={"option": "usernoread"}, ack=False
    )

    # get row indices relative to full main table
    if rowidxs is None:
        taql_rowid = f"select rowid() as ROWS from $mtable {taql_where}"
        with open_query(mtable, taql_rowid) as query_rows:
            rowidxs = query_rows.getcol("ROWS")
            mtable.close()

    nrows = len(rowidxs)
    if nrows == 0:
        return xr.Dataset()

    part_ids = get_partition_ids(mtable, taql_where)

    taql_cols = f"select * from $mtable {taql_where}"
    with open_query(mtable, taql_cols) as query_cols:
        cols = query_cols.colnames()
        ignore = [
            col
            for col in cols
            if (not query_cols.iscelldefined(col, 0))
            or (query_cols.coldatatype(col) == "record")
        ]
        cdata = dict(
            [
                (col, query_cols.getcol(col, 0, 1))
                for col in cols
                if (col not in ignore)
                and (ignore_msv2_cols and (col not in ignore_msv2_cols))
            ]
        )
        chan_cnt, pol_cnt = [
            (cdata[cc].shape[1], cdata[cc].shape[2])
            for cc in cdata
            if len(cdata[cc].shape) == 3
        ][0]

    mtable.close()

    mvars, mcoords, bvars, xds = {}, {}, {}, xr.Dataset()
    # loop over row chunks
    for rc in range(0, nrows, chunks[0]):
        crlen = min(chunks[0], nrows - rc)  # chunk row length
        rcidxs = rowidxs[rc : rc + chunks[0]]

        # loop over each column and create delayed dask arrays
        for col in cdata.keys():
            if col not in bvars:
                bvars[col] = []

            if len(cdata[col].shape) == 1:
                delayed_array = dask.delayed(read_flat_col_chunk)(
                    infile, col, (crlen,), rcidxs, None, None
                )
                bvars[col] += [
                    dask.array.from_delayed(delayed_array, (crlen,), cdata[col].dtype)
                ]

            elif col == "UVW":
                delayed_array = dask.delayed(read_flat_col_chunk)(
                    infile, col, (crlen, 3), rcidxs, None, None
                )
                bvars[col] += [
                    dask.array.from_delayed(delayed_array, (crlen, 3), cdata[col].dtype)
                ]

            elif len(cdata[col].shape) == 2:
                pol_list = []
                dd = 1 if cdata[col].shape[1] == chan_cnt else 2
                for pc in range(0, cdata[col].shape[1], chunks[dd]):
                    plen = min(chunks[dd], cdata[col].shape[1] - pc)
                    delayed_array = dask.delayed(read_flat_col_chunk)(
                        infile, col, (crlen, plen), rcidxs, None, pc
                    )
                    pol_list += [
                        dask.array.from_delayed(
                            delayed_array, (crlen, plen), cdata[col].dtype
                        )
                    ]
                bvars[col] += [dask.array.concatenate(pol_list, axis=1)]

            elif len(cdata[col].shape) == 3:
                chan_list = []
                for cc in range(0, chan_cnt, chunks[1]):
                    clen = min(chunks[1], chan_cnt - cc)
                    pol_list = []
                    for pc in range(0, cdata[col].shape[2], chunks[2]):
                        plen = min(chunks[2], cdata[col].shape[2] - pc)
                        delayed_array = dask.delayed(read_flat_col_chunk)(
                            infile, col, (crlen, clen, plen), rcidxs, cc, pc
                        )
                        pol_list += [
                            dask.array.from_delayed(
                                delayed_array, (crlen, clen, plen), cdata[col].dtype
                            )
                        ]
                    chan_list += [dask.array.concatenate(pol_list, axis=2)]
                bvars[col] += [dask.array.concatenate(chan_list, axis=1)]

    # now concat all the dask chunks from each time to make the xds
    mvars = {}
    for kk in bvars.keys():
        # from uppercase MS col names to lowercase xds var names:
        data_var = kk.lower()
        if len(bvars[kk]) == 0:
            ignore += [kk]
            continue
        if kk == "UVW":
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "uvw_coords"]
            )
        elif len(bvars[kk][0].shape) == 2 and (bvars[kk][0].shape[-1] == pol_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "pol"]
            )
        elif len(bvars[kk][0].shape) == 2 and (bvars[kk][0].shape[-1] == chan_cnt):
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0), dims=["row", "chan"]
            )
        else:
            mvars[data_var] = xr.DataArray(
                dask.array.concatenate(bvars[kk], axis=0),
                dims=["row", "freq", "pol"][: len(bvars[kk][0].shape)],
            )

    mvars["time"] = xr.DataArray(
        convert_casacore_time(mvars["TIME"].values), dims=["row"]
    ).chunk({"row": chunks[0]})

    # add xds global attributes
    cc_attrs = extract_table_attributes(infile)
    attrs = {"other": {"msv2": {"ctds_attrs": cc_attrs, "bad_cols": ignore}}}
    # add per data var attributes
    mvars = add_units_measures(mvars, cc_attrs)
    mcoords = add_units_measures(mcoords, cc_attrs)

    mvars = rename_vars(mvars)
    mvars = redim_id_data_vars(mvars)
    xds = xr.Dataset(mvars, coords=mcoords)

    return xds, part_ids, attrs



col_to_data_variable_names = {
    "FLOAT_DATA": "SPECTRUM",
    "DATA": "VISIBILITY",
    "CORRECTED_DATA": "VISIBILITY_CORRECTED",
    "WEIGHT_SPECTRUM": "WEIGHT",
    "WEIGHT": "WEIGHT",
    "FLAG": "FLAG",
    "UVW": "UVW",
    "TIME_CENTROID": "TIME_CENTROID",
    "EXPOSURE": "EFFECTIVE_INTEGRATION_TIME",
}
col_dims = {
    "DATA": ("time", "baseline_id", "frequency", "polarization"),
    "CORRECTED_DATA": ("time", "baseline_id", "frequency", "polarization"),
    "WEIGHT_SPECTRUM": ("time", "baseline_id", "frequency", "polarization"),
    "WEIGHT": ("time", "baseline_id", "frequency", "polarization"),
    "FLAG": ("time", "baseline_id", "frequency", "polarization"),
    "UVW": ("time", "baseline_id", "uvw_label"),
    "TIME_CENTROID": ("time", "baseline_id"),
    "EXPOSURE": ("time", "baseline_id"),
    "FLOAT_DATA": ("time", "baseline_id", "frequency", "polarization"),
}
col_to_coord_names = {
    "TIME": "time",
    "ANTENNA1": "baseline_ant1_id",
    "ANTENNA2": "baseline_ant2_id",
}

casacore_to_msv4_measure_type = {
    "quanta": {"type": "quantity", "Ref": None},
    "direction": {"type": "sky_coord", "Ref": "frame"},
    "epoch": {"type": "time", "Ref": "scale"},
    "frequency": {"type": "spectral_coord", "Ref": "frame"},
    "position": {"type": "earth_location", "Ref": "ellipsoid"},
    "uvw": {"type": "uvw", "Ref": "frame"},
}

casacore_to_msv4_ref = {"J2000": "FK5", "ITRF": "GRS80"}


def column_description_casacore_to_msv4_measure(
    casacore_column_description, ref_code=None, time_format="unix"
):
    import numpy as np

    msv4_measure = {}
    if "MEASINFO" in casacore_column_description["keywords"]:
        msv4_measure["type"] = casacore_to_msv4_measure_type[
            casacore_column_description["keywords"]["MEASINFO"]["type"]
        ]["type"]
        msv4_measure["units"] = list(
            casacore_column_description["keywords"]["QuantumUnits"]
        )

        if "TabRefCodes" in casacore_column_description["keywords"]["MEASINFO"]:
            ref_index = np.where(
                casacore_column_description["keywords"]["MEASINFO"]["TabRefCodes"]
                == ref_code
            )[0][0]
            casa_ref = casacore_column_description["keywords"]["MEASINFO"][
                "TabRefTypes"
            ][ref_index]
        else:
            casa_ref = casacore_column_description["keywords"]["MEASINFO"]["Ref"]

        if casa_ref in casacore_to_msv4_ref:
            casa_ref = casacore_to_msv4_ref[casa_ref]
        msv4_measure[
            casacore_to_msv4_measure_type[
                casacore_column_description["keywords"]["MEASINFO"]["type"]
            ]["Ref"]
        ] = casa_ref

        if msv4_measure["type"] == "time":
            msv4_measure["format"] = "unix"
    return msv4_measure


def create_attribute_metadata(col, main_column_descriptions):
    attrs_metadata = column_description_casacore_to_msv4_measure(
        main_column_descriptions[col]
    )
    if col in ["DATA", "CORRECTED_DATA", "WEIGHT"]:
        if not attrs_metadata:
            attrs_metadata["type"] = "quanta"
            attrs_metadata["units"] = ["unkown"]

    return attrs_metadata



import xarray as xr
import zarr
import copy
import os
from ._processing_set import processing_set

DIMENSION_KEY = "_ARRAY_DIMENSIONS"  # Used by xarray to store array labeling info in zarr meta data.
# from xradio._utils._logger import _get_logger


def _get_attrs(zarr_obj):
    """
    get attributes of zarr obj (groups or arrays)
    """
    return {k: v for k, v in zarr_obj.attrs.asdict().items() if not k.startswith("_NC")}


def _load_ms_xds(
    ps_name, ms_xds_name, slice_dict={}, cache_dir=None, chunk_id=None, date_time=""
):
    # logger = _get_logger()
    if cache_dir:
        xds_cached_name = (
            os.path.join(cache_dir, ms_xds_name) + "_" + str(chunk_id) + "_" + date_time
        )

        # Check if already chached:
        try:
            ms_xds = _load_ms_xds_core(
                ms_xds_name=xds_cached_name, slice_dict=slice_dict
            )

            # logger.debug(ms_xds_name + ' chunk ' + str(slice_dict) + ' was found in cache: ' + xds_cached)
            found_in_cache = True
            return xds, found_in_cache
        except:
            # logger.debug(xds_cached + ' chunk ' + str(slice_dict) + ' was not found in cache or failed to load. Retrieving chunk from ' + ms_xds_name + ' .')
            ms_xds = _load_ms_xds_core(
                ms_xds_name=os.path.join(ps_name, ms_xds_name), slice_dict=slice_dict
            )
            write_ms_xds(ms_xds, xds_cached_name)

            found_in_cache = False
            return xds, found_in_cache
    else:
        found_in_cache = None
        ms_xds = _load_ms_xds_core(
            ms_xds_name=os.path.join(ps_name, ms_xds_name), slice_dict=slice_dict
        )
        return ms_xds, found_in_cache


def _write_ms_xds(ms_xds, ms_xds_name):
    ms_xds_temp = ms_xds
    xr.Dataset.to_zarr(
        ms_xds.attrs["ANTENNA"],
        os.path.join(xds_cached_name, "ANTENNA"),
        consolidated=True,
    )
    ms_xds_temp = ms_xds
    ms_xds_temp.attrs["ANTENNA"] = {}
    xr.Dataset.to_zarr(
        ms_xds_temp, os.path.join(xds_cached_name, "MAIN"), consolidated=True
    )


def _load_ms_xds_core(ms_xds_name, slice_dict):
    ms_xds = _load_no_dask_zarr(
        zarr_name=os.path.join(ms_xds_name, "MAIN"), slice_dict=slice_dict
    )
    ms_xds.attrs["antenna_xds"] = _load_no_dask_zarr(
        zarr_name=os.path.join(ms_xds_name, "ANTENNA")
    )
    return ms_xds


def _load_no_dask_zarr(zarr_name, slice_dict={}):
    """
    Alternative to xarray open_zarr where the arrays are not Dask Arrays.

    slice_dict: A dictionary of slice objects for which values to read form a dimension.
                For example silce_dict={'time':slice(0,10)} would select the first 10 elements in the time dimension.
                If a dim is not specified all values are retruned.
    return:
        xarray.Dataset()

    #Should go into general utils.
    """

    # logger = _get_logger()
    zarr_group = zarr.open_group(store=zarr_name, mode="r")
    group_attrs = _get_attrs(zarr_group)

    slice_dict_complete = copy.deepcopy(slice_dict)
    coords = {}
    xds = xr.Dataset()
    for var_name, var in zarr_group.arrays():
        var_attrs = _get_attrs(var)

        for dim in var_attrs[DIMENSION_KEY]:
            if dim not in slice_dict_complete:
                slice_dict_complete[dim] = slice(None)  # No slicing.

        if (var_attrs[DIMENSION_KEY][0] == var_name) and (
            len(var_attrs[DIMENSION_KEY]) == 1
        ):
            coords[var_name] = var[
                slice_dict_complete[var_attrs[DIMENSION_KEY][0]]
            ]  # Dimension coordinates.
        else:
            # Construct slicing
            slicing_list = []
            for dim in var_attrs[DIMENSION_KEY]:
                slicing_list.append(slice_dict_complete[dim])
            slicing_tuple = tuple(slicing_list)
            xds[var_name] = xr.DataArray(
                var[slicing_tuple], dims=var_attrs[DIMENSION_KEY]
            )

    xds = xds.assign_coords(coords)

    xds.attrs = group_attrs

    return xds


def load_processing_set(ps_name, sel_parms):
    """
    sel_parms
        A dictionary where the keys are the names of the ms_xds's and the values are slice_dicts.
        slice_dicts: A dictionary where the keys are the dimension names and the values are slices.
    """
    ps = processing_set()
    for name_ms_xds, ms_xds_sel_parms in sel_parms.items():
        ps[name_ms_xds] = _load_ms_xds(ps_name, name_ms_xds, ms_xds_sel_parms)[0]
    return ps



import os
import xarray as xr
import pandas as pd
from ._processing_set import processing_set


def read_processing_set(ps_name, intents=None, fields=None):
    items = os.listdir(ps_name)
    ms_xds = xr.Dataset()
    ps = processing_set()
    for i in items:
        if "ddi" in i:
            xds = xr.open_zarr(ps_name + "/" + i + "/MAIN")

            if (intents is None) or (xds.attrs["intent"] in intents):
                if (fields is None) or (xds.attrs["field_info"]["name"] in fields):
                    ps[i] = xds
                    ps[i].attrs["antenna_xds"] = xr.open_zarr(
                        ps_name + "/" + i + "/ANTENNA"
                    )

    return ps



from importlib_metadata import version
import logging, multiprocessing, psutil
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import xarray as xr

from .cds import CASAVisSet
from .stokes_types import stokes_types


def make_coords(
    xds: xr.Dataset, ddi: int, subtables: Tuple[xr.Dataset, ...]
) -> Dict[str, np.ndarray]:
    """Make the coords to be added to a partition or chunk (besides
    the time, baseline) basic structure

    Grabs:
    - channel (center) frequency values from the spw subtable
    - pol idxs from the pol+ddi subtables -> pol names via the stokes_types
    - antenna IDs from antenna subtable

    :param: sub-xds as (ant_xds, ddi_xds, spw_xds, pol_xds)
    """
    ant_xds, ddi_xds, spw_xds, pol_xds = subtables
    freq = spw_xds.chan_freq.values[
        ddi_xds.spectral_window_id.values[ddi], : xds.freq.shape[0]
    ]
    pol_ids = pol_xds.corr_type.values[
        ddi_xds.polarization_id.values[ddi], : xds.pol.shape[0]
    ]
    pol_names = np.vectorize(stokes_types.get)(pol_ids)
    ant_id = ant_xds.antenna_id.values
    coords = {
        "freq": freq,
        "pol": pol_names,
        "antenna_id": ant_id
        # These will be metainfo in partitions
        # "spw_id": [ddi_xds["spectral_window_id"].values[ddi]],
        # "pol_id": [ddi_xds["polarization_id"].values[ddi]],
    }
    return coords


def vis_xds_packager_cds(
    subtables: List[Tuple[str, xr.Dataset]],
    partitions: Dict[Any, xr.Dataset],
    descr_add: str = "",
):
    """Takes a a list of subtable xds datasets and a dictionary of data
    partition xds datasets and and packages them as a CASA vis dataset
    (cds)

    :param partitions: data partiions as xds datasets
    :param subtables: subtables as xds datasets
    :param descr_add: substring to add to the short descr string of the cds
    :return: A "cds" - container for the metainfo subtables and data partitions
    """
    vers = version("xradio")

    return CASAVisSet(
        subtables,
        partitions,
        f"CASA vis set produced by xradio {vers}/{descr_add}",
    )


def vis_xds_packager_mxds(
    partitions: Dict[Any, xr.Dataset],
    subtables: List[Tuple[str, xr.Dataset]],
    add_global_coords: bool = True,
) -> xr.Dataset:
    """Takes a dictionary of data partition xds datasets and a list of
    subtable xds datasets and packages them as a dataset of datasets
    (mxds)

    :param partitions: data partiions as xds datasets
    :param subtables: subtables as xds datasets
    :add_global_coords: whether to add coords to the output mxds
    :return: A "mxds" - xr.dataset of datasets
    """
    mxds = xr.Dataset(attrs={"metainfo": subtables, "partitions": partitions})

    if add_global_coords:
        mxds = mxds.assign_coords(make_global_coords(mxds))

    return mxds


def make_global_coords(mxds: xr.Dataset):
    coords = {}
    metainfo = mxds.attrs["metainfo"]
    if "antenna" in metainfo:
        coords["antenna_ids"] = metainfo["antenna"].antenna_id.values
        coords["antennas"] = xr.DataArray(
            metainfo["antenna"].name.values, dims=["antenna_ids"]
        )
    if "field" in metainfo:
        coords["field_ids"] = metainfo["field"].field_id.values
        coords["fields"] = xr.DataArray(
            metainfo["field"].name.values, dims=["field_ids"]
        )
    if "feed" in mxds.attrs:
        coords["feed_ids"] = metainfo["feed"].feed_id.values
    if "observation" in metainfo:
        coords["observation_ids"] = metainfo["observation"].observation_id.values
        coords["observations"] = xr.DataArray(
            metainfo["observation"].project.values, dims=["observation_ids"]
        )
    if "polarization" in metainfo:
        coords["polarization_ids"] = metainfo["polarization"].pol_setup_id.values
    if "source" in metainfo:
        coords["source_ids"] = metainfo["source"].source_id.values
        coords["sources"] = xr.DataArray(
            metainfo["source"].name.values, dims=["source_ids"]
        )
    if "spectral_window" in metainfo:
        coords["spw_ids"] = metainfo["spectral_window"].spw_id.values
    if "state" in metainfo:
        coords["state_ids"] = metainfo["state"].state_id.values

    return coords


def expand_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    expand single (row) dimension of xds to (time, baseline)
    """
    assert "baseline" not in xds.coords

    txds = xds.copy()
    unique_baselines, baselines = np.unique(
        [txds.antenna1.values, txds.antenna2.values], axis=1, return_inverse=True
    )
    txds["baseline"] = xr.DataArray(baselines.astype("int32"), dims=["row"])

    try:
        txds = (
            txds.set_index(row=["time", "baseline"])
            .unstack("row")
            .transpose("time", "baseline", ...)
        )

        # unstack changes type to float when it needs to introduce NaNs, so
        # we need to reset to the proper type. Avoid if possible, as the
        # astype are costly
        for dv in txds.data_vars:
            if txds[dv].dtype != xds[dv].dtype:
                txds[dv] = txds[dv].astype(xds[dv].dtype)
    except Exception as exc:
        print(
            f"WARNING: Cannot expand rows to (time, baseline), "
            f"possibly duplicate values in (time, baseline). Exception: {exc}"
        )
        txds = xds.copy()

    return txds


def flatten_xds(xds: xr.Dataset) -> xr.Dataset:
    """
    flatten (time, baseline) dimensions of xds back to single dimension (row)
    """
    nan_int = np.array([np.nan]).astype("int32")[0]
    txds = xds.copy()

    # flatten the time x baseline dimensions of main table
    if ("time" in xds.dims) and ("baseline" in xds.dims):
        txds = xds.stack({"row": ("time", "baseline")}).transpose("row", ...)
        txds = txds.where(
            (txds.state_id != nan_int) & (txds.field_id != nan_int), drop=True
        )  # .unify_chunks()
        for dv in list(xds.data_vars):
            txds[dv] = txds[dv].astype(xds[dv].dtype)

    return txds


####################################
# xautomatically compute best data chunking
def optimal_chunking(
    ndim: Union[int, None] = None,
    didxs: Union[Tuple[int], List[int], None] = None,
    chunk_size: str = "auto",
    data_shape: Union[tuple, None] = None,
) -> tuple:
    """
    Determine the optimal chunk shape for reading an MS or Image based
    on machine resources and intended operations

    :param ndim: number of dimensions to chunk. An MS is 3, an
    expanded MS is 4. An image could be anywhere from 2 to 5. Not
    needed if data_shape is given.
    :param didxs: dimension indices over which subsequent operations
    will be performed. Values should be less than ndim. Tries to
    reduce inter-process communication of data contents. Needs to
    know the shape to do this well. Default None balances chunk size
    across all dimensions.
    :param chunk_size: target chunk size ('large', 'small', 'auto').
    Default 'auto' tries to guess by looking at CPU core count and
    available memory.
    :param data_shape: shape of the total MS DDI or Image data. Helps
    to know. Default None does not optimize based on shape

    :return: optimal chunking for reading the ms (row, chan, pol)
    """
    assert (ndim is not None) or (
        data_shape is not None
    ), "either ndim or data_shape must be given"
    assert chunk_size in ["large", "small", "auto"], "invalid chunk_size parameter"
    if ndim is None:
        ndim = len(data_shape)

    opt_dims = (
        didxs if (didxs is not None) and (len(didxs) > 0) else np.arange(ndim)
    )  # maximize these dim chunk sizes
    nonopt_dims = np.setdiff1d(np.arange(ndim), opt_dims)  # at the expense of these

    max_chunk_sizes = (
        data_shape
        if data_shape is not None
        else [dd for ii, dd in enumerate([10000, 10000, 10000, 4, 10]) if ii < ndim]
    )
    min_chunk_sizes = (
        np.ceil(np.array(data_shape) / 80).astype(int)
        if data_shape is not None
        else [1000, 1, 1]
        if ndim == 3
        else [dd for ii, dd in enumerate([10, 10, 1, 1, 1]) if ii < ndim]
    )
    target_size = (
        175 * 1024**2 / 8
    )  # ~175 MB chunk worst case with 8-byte DATA column
    bytes_per_core = int(
        round(
            ((psutil.virtual_memory().available * 0.10) / multiprocessing.cpu_count())
        )
    )
    if data_shape is not None:
        bytes_per_core = min(
            bytes_per_core, np.prod(data_shape) * 8 / 2
        )  # ensure at least two chunks
    if chunk_size == "large":
        target_size = target_size * 6  # ~1 GB
    if chunk_size == "auto":
        target_size = max(min(target_size * 6, bytes_per_core / 8), target_size)

    # start by setting the optimized dims to their max size and non-optimized dims to their min size
    chunks = np.zeros((ndim), dtype="int")
    chunks[opt_dims] = np.array(max_chunk_sizes)[opt_dims]
    chunks[nonopt_dims] = np.array(min_chunk_sizes)[nonopt_dims]

    # iteratively walk towards an optimal chunk size
    # iteration is needed because rounding to nearest integer index can make a big different (2x) in chunk size
    # for small dimensions like pol
    for ii in range(10):
        # if the resulting size is too big, reduce the sizes of the optimized dimensions
        if (np.prod(chunks) > target_size) and (len(opt_dims) > 0):
            chunks[opt_dims] = np.round(
                chunks[opt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(opt_dims))
            )
        # else if the resulting size is too small, increase the sizes of the non-optimized dimensions
        elif (np.prod(chunks) < target_size) and (len(nonopt_dims) > 0):
            chunks[nonopt_dims] = np.round(
                chunks[nonopt_dims]
                * (target_size / np.prod(chunks)) ** (1 / len(nonopt_dims))
            )
        chunks = np.min((chunks, max_chunk_sizes), axis=0)
        chunks = np.max((chunks, min_chunk_sizes), axis=0)

    return tuple(chunks)


def calc_optimal_ms_chunk_shape(
    memory_available_in_bytes, shape, element_size_in_bytes, column_name
) -> int:
    """
    Calculates the max number of rows (1st dim in shape) of a variable
    that can be fit in the memory for a thread.
    """
    factor = 0.8  # Account for memory used by other objects in thread.
    # total_mem = np.prod(shape)*element_size_in_bytes
    single_row_mem = np.prod(shape[1:]) * element_size_in_bytes

    try:
        assert single_row_mem < factor * memory_available_in_bytes
    except AssertionError as err:
        logging.exception(
            "Not engough memory in a thread to contain a row of "
            + column_name
            + ". Need at least "
            + str(single_row_mem / factor)
            + " bytes."
        )
        raise err

    rows_chunk_size = int((factor * memory_available_in_bytes) / single_row_mem)

    if rows_chunk_size > shape[0]:
        rows_chunk_size = shape[0]

    logging.debug(
        "Numbers of rows in chunk for " + column_name + ": " + str(rows_chunk_size)
    )

    return rows_chunk_size



import astropy
from astropy.coordinates import Angle, SkyCoord
from casacore import images, tables
import copy
import dask.array as da
import numpy as np
import os
from typing import Union
import xarray as xr
from .common import __active_mask, __native_types, __object_name, __pointing_center
from ..common import __doppler_types
from ...._utils._casacore.tables import open_table_rw


# TODO move this to a common file to be shared
def __compute_ref_pix(xds: xr.Dataset, direction: dict) -> np.ndarray:
    # TODO more general coordinates
    long = xds.right_ascension
    lat = xds.declination
    ra_crval = long.attrs["wcs"]["crval"]
    dec_crval = lat.attrs["wcs"]["crval"]
    long_close = np.where(np.isclose(long, ra_crval))
    lat_close = np.where(np.isclose(lat, dec_crval))
    if long_close and lat_close:
        long_list = [(i, j) for i, j in zip(long_close[0], long_close[1])]
        lat_list = [(i, j) for i, j in zip(lat_close[0], lat_close[1])]
        common_indices = [t for t in long_list if t in lat_list]
        if len(common_indices) == 1:
            return np.array(common_indices[0])
    cdelt = max(abs(long.attrs["wcs"]["cdelt"]), abs(lat.attrs["wcs"]["cdelt"]))

    # this creates an image of mostly NaNs. The few pixels with values are
    # close to the reference pixel
    ra_diff = long - ra_crval
    dec_diff = lat - dec_crval
    # this returns a 2-tuple of indices where the values in aa are not NaN
    indices_close = np.where(
        ra_diff * ra_diff + dec_diff * dec_diff < 2 * cdelt * cdelt
    )
    # this determines the closest pixel to the reference pixel
    closest = 5e10
    pix = []
    for i, j in zip(indices_close[0], indices_close[1]):
        dra = long[i, j] - ra_crval
        ddec = lat[i, j] - dec_crval
        if dra * dra + ddec * ddec < closest:
            pix = [i, j]
    xds_dir = xds.attrs["direction"]
    # get the actual ref pix
    proj = direction["projection"]
    wcs_dict = {}
    wcs_dict[f"CTYPE1"] = f"RA---{proj}"
    wcs_dict[f"NAXIS1"] = long.shape[0]
    wcs_dict[f"CUNIT1"] = long.attrs["unit"]
    # FITS arrays are 1-based
    wcs_dict[f"CRPIX1"] = pix[0] + 1
    wcs_dict[f"CRVAL1"] = long[pix[0], pix[1]].item(0)
    wcs_dict[f"CDELT1"] = long.attrs["wcs"]["cdelt"]
    wcs_dict[f"CTYPE2"] = f"DEC--{proj}"
    wcs_dict[f"NAXIS2"] = lat.shape[1]
    wcs_dict[f"CUNIT2"] = lat.attrs["unit"]
    # FITS arrays are 1-based
    wcs_dict[f"CRPIX2"] = pix[1] + 1
    wcs_dict[f"CRVAL2"] = lat[pix[0], pix[1]].item(0)
    wcs_dict[f"CDELT2"] = lat.attrs["wcs"]["cdelt"]
    w = astropy.wcs.WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    sky = SkyCoord(
        ra_crval,
        dec_crval,
        frame=xds.attrs["direction"]["system"].lower(),
        equinox=xds.attrs["direction"]["equinox"],
        unit=long.attrs["unit"],
    )
    return w.world_to_pixel(sky)


def __compute_direction_dict(xds: xr.Dataset) -> dict:
    """
    Given xds metadata, compute the direction dict that is valid
    for a CASA image coordinate system
    """
    direction = {}
    xds_dir = xds.attrs["direction"]
    direction["system"] = xds_dir["equinox"]
    direction["projection"] = xds_dir["projection"]
    direction["projection_parameters"] = xds_dir["projection_parameters"]
    long = xds.right_ascension
    lat = xds.declination
    direction["units"] = np.array([long.attrs["unit"], lat.attrs["unit"]], dtype="<U16")
    direction["crval"] = np.array(
        [long.attrs["wcs"]["crval"], lat.attrs["wcs"]["crval"]]
    )
    direction["cdelt"] = np.array(
        [long.attrs["wcs"]["cdelt"], lat.attrs["wcs"]["cdelt"]]
    )
    """
    # get the actual ref pix
    proj = direction['projection']
    wcs_dict = {}
    wcs_dict[f'CTYPE1'] = f'RA---{proj}'
    wcs_dict[f'NAXIS1'] = long.shape[0]
    wcs_dict[f'CUNIT1'] = long.attrs['unit']
    # FITS arrays are 1-based
    wcs_dict[f'CRPIX1'] = 1
    wcs_dict[f'CRVAL1'] = long[0][0].item(0)
    wcs_dict[f'CDELT1'] = long.attrs['wcs']['cdelt']
    wcs_dict[f'CTYPE2'] = f'DEC--{proj}'
    wcs_dict[f'NAXIS2'] = lat.shape[1]
    wcs_dict[f'CUNIT2'] = lat.attrs['unit']
    # FITS arrays are 1-based
    wcs_dict[f'CRPIX2'] = 1
    wcs_dict[f'CRVAL2'] = lat[0][0].item(0)
    wcs_dict[f'CDELT2'] = lat.attrs['wcs']['cdelt']
    print('*** wcs_dict', wcs_dict)
    w = astropy.wcs.WCS(wcs_dict)
    x, y = np.indices(w.pixel_shape)
    sky = SkyCoord(direction[
        'crval'][0], direction['crval'][1],
        frame=xds.attrs['direction']['system'].lower(),
        equinox=xds.attrs['direction']['equinox'],
        unit=long.attrs['unit']
    )
    crpix = w.world_to_pixel(sky)
    """
    crpix = __compute_ref_pix(xds, direction)
    direction["crpix"] = np.array([crpix[0], crpix[1]])
    direction["pc"] = xds_dir["pc"]
    direction["axes"] = ["Right Ascension", "Declination"]
    direction["conversionSystem"] = direction["system"]
    for s in ["longpole", "latpole"]:
        direction[s] = Angle(str(xds_dir[s]["value"]) + xds_dir[s]["unit"]).deg
    return direction


def __compute_spectral_dict(
    xds: xr.Dataset,
    direction: dict,
    obsdate: dict,
    tel_pos: dict,
) -> dict:
    """
    Given xds metadata, compute the spectral dict that is valid
    for a CASA image coordinate system
    """
    spec = {}
    spec_conv = copy.deepcopy(xds.frequency.attrs["conversion"])
    for k in ("direction", "epoch", "position"):
        spec_conv[k]["type"] = k
    spec_conv["direction"]["refer"] = spec_conv["direction"]["system"]
    del spec_conv["direction"]["system"]
    if (
        spec_conv["direction"]["refer"] == "FK5"
        and spec_conv["direction"]["equinox"] == "J2000"
    ):
        spec_conv["direction"]["refer"] = "J2000"
    del spec_conv["direction"]["equinox"]
    spec["conversion"] = spec_conv
    spec["formatUnit"] = ""
    spec["name"] = "Frequency"
    spec["nativeType"] = __native_types.index(xds.frequency.attrs["native_type"])
    spec["restfreq"] = xds.frequency.attrs["restfreq"]
    spec["restfreqs"] = copy.deepcopy(xds.frequency.attrs["restfreqs"])
    spec["system"] = xds.frequency.attrs["system"]
    spec["unit"] = xds.frequency.attrs["unit"]
    spec["velType"] = __doppler_types.index(xds.velocity.attrs["doppler_type"])
    spec["velUnit"] = xds.velocity.attrs["unit"]
    spec["version"] = 2
    spec["waveUnit"] = xds.frequency.attrs["wave_unit"]
    spec_wcs = copy.deepcopy(xds.frequency.attrs["wcs"])
    spec_wcs["ctype"] = "FREQ"
    spec_wcs["pc"] = 1.0
    spec_wcs["crpix"] = (spec_wcs["crval"] - xds.frequency.values[0]) / spec_wcs[
        "cdelt"
    ]
    spec["wcs"] = spec_wcs
    return spec


def __coord_dict_from_xds(xds: xr.Dataset) -> dict:
    coord = {}
    coord["telescope"] = xds.attrs["telescope"]["name"]
    coord["observer"] = xds.attrs["observer"]
    obsdate = {}
    obsdate["refer"] = xds.coords["time"].attrs["time_scale"]
    obsdate["type"] = "epoch"
    obsdate["m0"] = {}
    obsdate["m0"]["unit"] = xds.coords["time"].attrs["unit"]
    obsdate["m0"]["value"] = xds.coords["time"].values[0]
    # obsdate['format'] = xds.time.attrs['format']
    coord["obsdate"] = obsdate
    coord["pointingcenter"] = xds.attrs[__pointing_center].copy()
    coord["telescopeposition"] = xds.attrs["telescope"]["position"].copy()
    coord["direction0"] = __compute_direction_dict(xds)
    coord["stokes1"] = {
        "axes": np.array(["Stokes"], dtype="<U16"),
        "cdelt": np.array([1.0]),
        "crpix": np.array([0.0]),
        "crval": np.array([1.0]),
        "pc": np.array([[1.0]]),
        "stokes": np.array(xds.polarization.values, dtype="<U16"),
    }
    coord["spectral2"] = __compute_spectral_dict(
        xds, coord["direction0"], coord["obsdate"], coord["telescopeposition"]
    )
    coord["pixelmap0"] = np.array([0, 1])
    coord["pixelmap1"] = np.array([2])
    coord["pixelmap2"] = np.array([3])
    coord["pixelreplace0"] = np.array([0.0, 0.0])
    coord["pixelreplace1"] = np.array([0.0])
    coord["pixelreplace2"] = np.array([0.0])
    coord["worldmap0"] = np.array([0, 1], dtype=np.int32)
    coord["worldmap1"] = np.array([2], dtype=np.int32)
    coord["worldmap2"] = np.array([3], dtype=np.int32)
    # coord['worldreplace0'] = coord['direction0']['crval']
    # this probbably needs some verification
    coord["worldreplace0"] = [0.0, 0.0]
    coord["worldreplace1"] = np.array(coord["stokes1"]["crval"])
    coord["worldreplace2"] = np.array([xds.frequency.attrs["wcs"]["crval"]])
    return coord


def __history_from_xds(xds: xr.Dataset, image: str) -> None:
    nrows = len(xds.history.row) if "row" in xds.data_vars else 0
    if nrows > 0:
        # TODO need to implement nrows == 0 case
        with open_table_rw(os.sep.join([image, "logtable"])) as tb:
            tb.addrows(nrows + 1)
            for c in ["TIME", "PRIORITY", "MESSAGE", "LOCATION", "OBJECT_ID"]:
                vals = xds.history[c].values
                if c == "TIME":
                    k = time.time() + 40587 * 86400
                elif c == "PRIORITY":
                    k = "INFO"
                elif c == "MESSAGE":
                    k = (
                        "Wrote xds to "
                        + os.path.basename(image)
                        + " using cngi_io.xds_to_casa_image_2()"
                    )
                elif c == "LOCATION":
                    k = "cngi_io.xds_to_casa_image_2"
                elif c == "OBJECT_ID":
                    k = ""
                vals = np.append(vals, k)
                tb.putcol(c, vals)


def __imageinfo_dict_from_xds(xds: xr.Dataset) -> dict:
    ii = {}
    ii["image_type"] = (
        xds.sky.attrs["image_type"] if "image_type" in xds.sky.attrs else ""
    )
    ii["objectname"] = xds.attrs[__object_name]
    if "beam" in xds.data_vars:
        # multi beam
        pp = {}
        pp["nChannels"] = len(xds.frequency)
        pp["nStokes"] = len(xds.polarization)
        bu = xds.beam.attrs["unit"]
        chan = 0
        polarization = 0
        bv = xds.beam.values
        for i in range(pp["nChannels"] * pp["nStokes"]):
            bp = bv[0][polarization][chan][:]
            b = {
                "major": {"unit": bu, "value": bp[0]},
                "minor": {"unit": bu, "value": bp[1]},
                "positionangle": {"unit": bu, "value": bp[2]},
            }
            pp["*" + str(pp["nChannels"] * polarization + chan)] = b
            chan += 1
            if chan >= pp["nChannels"]:
                chan = 0
                polarization += 1
        ii["perplanebeams"] = pp
    elif "beam" in xds.attrs and xds.attrs["beam"]:
        # do nothing if xds.attrs['beam'] is None
        ii["restoringbeam"] = xds.attrs["beam"]
    return ii


def __write_casa_data(xds: xr.Dataset, image_full_path: str) -> None:
    sky_ap = "sky" if "sky" in xds else "apeature"
    if xds[sky_ap].shape[0] != 1:
        raise Exception("XDS can only be converted if it has exactly one time plane")
    casa_image_shape = (
        xds[sky_ap]
        .isel(time=0)
        .transpose(*("frequency", "polarization", "m", "l"))
        .shape[::-1]
    )
    active_mask = xds.attrs["active_mask"] if __active_mask in xds.attrs else ""
    masks = []
    masks_rec = {}
    mask_rec = {
        "box": {
            "blc": np.array([1.0, 1.0, 1.0, 1.0]),
            "comment": "",
            "isRegion": 1,
            "name": "LCBox",
            "oneRel": True,
            "shape": np.array(casa_image_shape),
            "trc": np.array(casa_image_shape),
        },
        "comment": "",
        "isRegion": 1,
        "name": "LCPagedMask",
    }
    for m in xds.data_vars:
        attrs = xds[m].attrs
        if "image_type" in attrs and attrs["image_type"] == "Mask":
            masks_rec[m] = mask_rec
            masks_rec[m]["mask"] = f"Table: {os.sep.join([image_full_path, m])}"
            masks.append(m)
    myvars = [sky_ap]
    myvars.extend(masks)
    # xr.apply_ufunc seems to like stripping attributes from xds and its coordinates,
    # so make a deep copy of the pertinent thing to be handled in xr.apply_ufunc()
    nan_mask = xr.apply_ufunc(da.isnan, xds[sky_ap].copy(deep=True), dask="allowed")
    # test if nan pixels are all already masked in active_mask.
    # if so, don't worry about masking them again, just use active_mask
    # which already masks all nans
    arr_masks = {}
    do_mask_nans = False
    there_are_nans = nan_mask.any()
    if there_are_nans:
        has_active_mask = bool(active_mask)
        if not has_active_mask:
            do_mask_nans = True
        else:
            notted_active_mask = xr.apply_ufunc(
                da.logical_not, xds[active_mask].copy(deep=True), dask="allowed"
            )
            some_nans_are_not_already_masked = xr.apply_ufunc(
                da.logical_and, nan_mask, notted_active_mask, dask="allowed"
            )
            do_mask_nans = some_nans_are_not_already_masked.any()
    if do_mask_nans:
        mask_name = "mask_xds_nans"
        i = 0
        while mask_name in masks:
            mask_name = f"mask_xds_nans{i}"
            i += 1
        masks_rec[mask_name] = mask_rec
        masks_rec[mask_name][
            "mask"
        ] = f"Table: {os.sep.join([image_full_path, mask_name])}"
        masks.append(mask_name)
        arr_masks[mask_name] = nan_mask
        if active_mask:
            mask_name = f"mask_xds_nans_or_{active_mask}"
            i = 0
            while mask_name in masks:
                mask_name = f"mask_xds_nans{i}"
                i += 1
            masks_rec[mask_name] = mask_rec
            masks_rec[mask_name][
                "mask"
            ] = f"Table: {os.sep.join([image_full_path, mask_name])}"
            masks.append(mask_name)
            # This command strips attributes at various places for no
            # apparent reason, so make a copy of the xds to run it on
            arr_masks[mask_name] = xr.apply_ufunc(
                da.logical_or,
                xds[active_mask].copy(deep=True),
                nan_mask,
                dask="allowed",
            )
        active_mask = mask_name
    __write_initial_image(xds, image_full_path, active_mask, casa_image_shape[::-1])
    for v in myvars:
        __write_pixels(v, active_mask, image_full_path, xds)
    for name, v in arr_masks.items():
        __write_pixels(name, active_mask, image_full_path, xds, v)
    if masks:
        with open_table_rw(image_full_path) as tb:
            tb.putkeyword("masks", masks_rec)
            tb.putkeyword("Image_defaultmask", active_mask)


def __write_initial_image(
    xds: xr.Dataset, imagename: str, maskname: str, image_shape: tuple
):
    image_full_path = os.path.expanduser(imagename)
    # create the image and then delete the object
    if not maskname:
        maskname = ""
    casa_image = images.image(image_full_path, maskname=maskname, shape=image_shape)
    del casa_image


def __write_image_block(xda: xr.DataArray, outfile: str, blc: tuple) -> None:
    """
    Write image xda chunk to the corresponding image table slice
    """
    # trigger the DAG for this chunk and return values while the table is
    # unlocked
    values = xda.compute().values
    with open_table_rw(outfile) as tb_tool:
        tb_tool.putcellslice(
            tb_tool.colnames()[0],
            0,
            values,
            blc,
            tuple(np.array(blc) + np.array(values.shape) - 1),
        )


def __write_pixels(
    v: str,
    active_mask: str,
    image_full_path: str,
    xds: xr.Dataset,
    value: xr.DataArray = None,
) -> None:
    flip = False
    if v == "sky" or v == "apeture":
        filename = image_full_path
    else:
        # mask
        flip = True
        filename = os.sep.join([image_full_path, v])
        # the default mask has already been written in __xds_to_casa_image()
        if not os.path.exists(filename):
            tb = tables.table(os.sep.join([image_full_path, active_mask]))
            tb.copy(filename, deep=True, valuecopy=True)
            tb.close()
    arr = xds[v] if v in xds.data_vars else value
    arr = arr.isel(time=0).transpose(*("frequency", "polarization", "m", "l"))
    chunk_bounds = arr.chunks
    b = [0, 0, 0, 0]
    loc0, loc1, loc2, loc3 = (0, 0, 0, 0)
    for i0 in chunk_bounds[0]:
        b[0] = loc0
        s0 = slice(b[0], b[0] + i0)
        loc1 = 0
        for i1 in chunk_bounds[1]:
            b[1] = loc1
            s1 = slice(b[1], b[1] + i1)
            loc2 = 0
            for i2 in chunk_bounds[2]:
                b[2] = loc2
                s2 = slice(b[2], b[2] + i2)
                loc3 = 0
                for i3 in chunk_bounds[3]:
                    b[3] = loc3
                    blc = tuple(b)
                    s3 = slice(b[3], b[3] + i3)
                    sub_arr = arr[s0, s1, s2, s3]
                    if flip:
                        sub_arr = np.logical_not(sub_arr)
                    __write_image_block(sub_arr, filename, blc)
                    loc3 += i3
                loc2 += i2
            loc1 += i1
        loc0 += i0
